///////////////////////////////////////////////////////////////////////
// contain.c - EZTWAIN library, container handling functions
// Copyright (C) 1999-2001 by Dosadi.  All Rights Reserved.
//
// Download from:    www.dosadi.com
//
// Support contact:  support@dosadi.com
//
// See end of eztwain.c for revision history
//
///////////////////////////////////////////////////////////////////////
//------------ Includes

#include <assert.h>
#include <memory.h>		// memset, memcpy
#include <stdio.h>      // sprintf
#include <io.h>         // for _lseek
#include <crtdbg.h>     // _CrtIsValidPointer
#include <time.h>
#include "windows.h"

#ifdef _WIN32
#ifndef _INC_VFW
#include <vfw.h>
#endif
#else
#include "commdlg.h"
#endif

#include "twain.h"

#ifdef _WIN32
  #define EZTAPI WINAPI
#else
  #define EZTAPI __export FAR PASCAL
#endif

#include "eztwain.h"

#ifdef _WIN32
  #define INT32 int
  #define DATAGROUP unsigned
  #define DSM_FILENAME "TWAIN_32.DLL"
  #define DSM_ENTRYPOINT "DSM_Entry"
  #define IsValidHandle(h) (h!=NULL)
  #define HUGEWRITE(fh, pb, bc) ((INT32)_lwrite(fh, pb, bc))
  #define HUGEREAD(fh, pb, bc) ((INT32)_lread(fh, pb, bc))
  #define FMEMSET(p, v, n) memset(p, v, n);
  #define FMEMCPY(p, q, n) memcpy(p, q, n);
#else
  #define INT32 long
  #define DATAGROUP unsigned long
  #define DSM_FILENAME "TWAIN.DLL"
  #define DSM_ENTRYPOINT "DSM_ENTRY"
  #define VALID_HANDLE 32
  #define IsValidHandle(h) ((h!=0) && ((h) >= 32))
  #define HUGEWRITE(fh, pb, bc) _hwrite(fh, pb, bc)
  #define HUGEREAD(fh, pb, bc) _hread(fh, pb, bc)
  #define FMEMSET(p, v, n) _fmemset(p, v, n);
  #define FMEMCPY(p, q, n) _fmemcpy(p, q, n);
#endif


// Privately shared functions
extern double Fix32ToFloat(TW_FIX32 fix);
extern void SetFix32(TW_FIX32* pfix, double r);
extern const size_t nTypeSize[];
extern long EZround(double d);


static int StoreItem(void* pv, int twty, double dV)
// Store the value dV into the item at pv which has type twty
{
    switch (twty) {
    case TWTY_INT8:
        *(TW_INT8*)pv = (TW_INT8)dV;
        break;
    case TWTY_INT16:
        *(TW_INT16*)pv = (TW_INT16)dV;
        break;
    case TWTY_INT32:
        *(TW_INT32*)pv = (TW_INT32)dV;
        break;
    case TWTY_UINT8:
        *(TW_UINT8*)pv = (TW_UINT8)dV;
        break;
    case TWTY_UINT16:
        *(TW_UINT16*)pv = (TW_UINT16)dV;
        break;
    case TWTY_UINT32:
        *(TW_UINT32*)pv = (TW_UINT32)dV;
        break;
    case TWTY_BOOL:
        *(TW_BOOL*)pv = (TW_BOOL)dV;
        break;
    case TWTY_FIX32:
        SetFix32(pv, dV);
        break;
    default:
        return FALSE;
    } // switch
    return TRUE;
} // StoreItem


static double Value2Float(TW_UINT32 u32, TW_UINT16 twty)
{
    switch (twty) {
    case TWTY_INT8:
        return (TW_INT8)u32;
    case TWTY_INT16:
        return (TW_INT16)u32;
    case TWTY_INT32:
        return (TW_INT32)u32;
    case TWTY_UINT8:
        return (TW_UINT8)u32;
    case TWTY_UINT16:
        return (TW_UINT16)u32;
    case TWTY_UINT32:
        return (TW_UINT32)u32;
    case TWTY_BOOL:
        return (TW_BOOL)u32;
    case TWTY_FIX32:
        return TWAIN_Fix32ToFloat(u32);
    default:
        return 0.0;
    } // switch
} // Value2Float


static TW_UINT32 Float2Value(double dV, int twty)
{
    switch (twty) {
    case TWTY_INT8:
        return (TW_INT8)dV;
    case TWTY_INT16:
        return (TW_INT16)dV;
    case TWTY_INT32:
        return (TW_INT32)dV;
    case TWTY_UINT8:
        return (TW_UINT8)dV;
    case TWTY_UINT16:
        return (TW_UINT16)dV;
    case TWTY_UINT32:
        return (TW_UINT32)dV;
    case TWTY_BOOL:
        return (TW_BOOL)dV;
    case TWTY_FIX32:
        return TWAIN_ToFix32(dV);
    default:
        return 0;
    } // switch
} // Float2Value

static double Item2Float(void* pv, TW_UINT16 twty)
{
    switch (twty) {
    case TWTY_INT8:
        return *(TW_INT8*)pv;
    case TWTY_INT16:
        return *(TW_INT16*)pv;
    case TWTY_INT32:
        return *(TW_INT32*)pv;
    case TWTY_UINT8:
        return *(TW_UINT8*)pv;
    case TWTY_UINT16:
        return *(TW_UINT16*)pv;
    case TWTY_UINT32:
        return *(TW_UINT32*)pv;
    case TWTY_BOOL:
        return *(TW_BOOL*)pv;
    case TWTY_FIX32:
        return Fix32ToFloat(*(TW_FIX32*)pv);
    default:
        return 0.0;
    } // switch
} // Item2Float



/////////////////////////////////////////////////////////////////////////////
// Container Functions
/////////////////////////////////////////////////////////////////////////////

typedef union {
    TW_ONEVALUE     onevalue;
    TW_RANGE        range;
    TW_ENUMERATION  enumeration;
    TW_ARRAY        array;
} TW_CONTAINER, *PTW_CONTAINER;

typedef struct {
    DWORD           dwSig;
    int             nFormat;
    HGLOBAL         hCon;
    PTW_CONTAINER   pCon;
} EZCONTAINER, *PEZCONTAINER;

#define CONTAINER_SIGNATURE 0xFADE3210

/////////////////////////////////////////////////////////////////////////////
// Container helpers

static int ItemCount(PEZCONTAINER pcon)
{
    PTW_CONTAINER p;
    p = pcon->pCon;
    switch (pcon->nFormat) {
    case CONTAINER_ONEVALUE:
        return 1;
    case CONTAINER_ENUMERATION:
        return p->enumeration.NumItems;
    case CONTAINER_ARRAY:
        return p->array.NumItems;
    case CONTAINER_RANGE:
        {
            TW_UINT16 twty = p->range.ItemType;
            double dMin = Value2Float(p->range.MinValue, twty);
            double dMax = Value2Float(p->range.MaxValue, twty);
            double dStep = Value2Float(p->range.StepSize, twty);
            double dCount = (dMax - dMin) / dStep;
            if (dCount > (int)dCount) {
                // for example, Min=0, Max=9, Step=2, dCount= 4.5
                dCount += 1.0;
            }
            return 1+(int)dCount;
        }
    default:
        return -1;
    } // switch
} // ItemCount


static HCONTAINER WrapContainer(int nFormat, HGLOBAL hcon)
{
	TW_CONTAINER*	pv = (TW_CONTAINER*)GlobalLock(hcon);
    PEZCONTAINER    pcon = (PEZCONTAINER)malloc(sizeof (EZCONTAINER));
    if (!hcon || !pv || !pcon) {
        if (!pv) {
		    TWAIN_ErrorBox("DataSource bug: bad Container handle.");
        }
        GlobalFree(hcon);
        free(pcon);
        return 0;
    }
    pcon->dwSig = CONTAINER_SIGNATURE;
    pcon->nFormat = nFormat;
    pcon->hCon = hcon;
    pcon->pCon = pv;
    return (HCONTAINER)pcon;
} // WrapContainer


static DWORD ContainerSize(int nFormat, unsigned twty, int nItems)
{
    DWORD dwSize;
    switch (nFormat) {
    case CONTAINER_ONEVALUE:
        dwSize = sizeof (TW_ONEVALUE);
        if (nTypeSize[twty] > sizeof (TW_UINT32)) {
            dwSize +=  nTypeSize[twty] - sizeof (TW_UINT32);
        }
        break;
    case CONTAINER_RANGE:
        dwSize = sizeof (TW_RANGE);
        break;
    case CONTAINER_ENUMERATION:
        // Note, this is slightly high because of the 'ItemList' field & padding
        dwSize = sizeof (TW_ENUMERATION) + nTypeSize[twty] * nItems;
        break;
    case CONTAINER_ARRAY:
        // Note, this is slightly high because of the 'ItemList' field & padding
        dwSize = sizeof (TW_ARRAY) + nTypeSize[twty] * nItems;
        break;
    default:
        dwSize = 0;
        break;
    } // switch
    return dwSize;
} // ContainerSize


void EZTAPI CONTAINER_Free(HCONTAINER hcon)
// Free the memory and resources of a capability container.
{
    PEZCONTAINER pcon = (PEZCONTAINER)hcon;
    if (pcon != NULL && pcon->dwSig == CONTAINER_SIGNATURE) {
        GlobalUnlock(pcon->hCon);
        GlobalFree(pcon->hCon);
        free(pcon);
    }
}

HCONTAINER EZTAPI CONTAINER_Copy(HCONTAINER hcon)
// Create and return a copy of a container.
{
    PEZCONTAINER pcon = (PEZCONTAINER)hcon;
    DWORD dwSize;
    HANDLE hnew;
    LPBYTE pnew;
    if (pcon == NULL || pcon->dwSig != CONTAINER_SIGNATURE) {
        return 0;
    }
    dwSize = GlobalSize(pcon->hCon);
    hnew = GlobalAlloc(GMEM_MOVEABLE, dwSize);
    if (!hnew) {
        return 0;
    }
    pnew = GlobalLock(hnew);
    FMEMCPY(pnew, pcon->pCon, dwSize);
    GlobalUnlock(hnew);
    return WrapContainer(pcon->nFormat, hnew);
} // CONTAINER_Copy


HCONTAINER EZTAPI CONTAINER_OneValue(int twty, double dVal)
{
    DWORD dwSize = ContainerSize(CONTAINER_ONEVALUE, twty, 1);
    HGLOBAL hcon = GlobalAlloc(GMEM_MOVEABLE, dwSize);
    if (hcon) {
        pTW_ONEVALUE p = (pTW_ONEVALUE)GlobalLock(hcon);
        p->ItemType = (TW_UINT16)twty;
        p->Item = Float2Value(dVal, twty);
        GlobalUnlock(hcon);
    }
    return WrapContainer(CONTAINER_ONEVALUE, hcon);
}


HCONTAINER EZTAPI CONTAINER_Range(int twty, double dMin, double dMax, double dStep)
{
    HGLOBAL hcon;
    if (twty >= TWTY_FRAME) {
        // Ranges of frames or strings are not supported.
        return 0;
    }
    hcon = GlobalAlloc(GMEM_MOVEABLE, sizeof (TW_RANGE));
    if (hcon) {
        pTW_RANGE p = (pTW_RANGE)GlobalLock(hcon);
        p->ItemType = (TW_UINT16)twty;
        p->MinValue = Float2Value(dMin, twty);
        p->MaxValue = Float2Value(dMax, twty);
        p->StepSize = Float2Value(dStep, twty);
        p->DefaultValue = Float2Value(dMin, twty);
        p->CurrentValue = Float2Value(dMin, twty);
        GlobalUnlock(hcon);
    }
    return WrapContainer(CONTAINER_RANGE, hcon);
}


HCONTAINER EZTAPI CONTAINER_Array(int twty, int nItems)
{
    DWORD dwSize = ContainerSize(CONTAINER_ARRAY, twty, nItems);
    HGLOBAL hcon = GlobalAlloc(GMEM_MOVEABLE, dwSize);
    if (hcon) {
        pTW_ARRAY p = (pTW_ARRAY)GlobalLock(hcon);
        p->ItemType = (TW_UINT16)twty;
        p->NumItems = nItems;
        GlobalUnlock(hcon);
    }
    return WrapContainer(CONTAINER_ARRAY, hcon);
}


HCONTAINER EZTAPI CONTAINER_Enumeration(int twty, int nItems)
{
    DWORD dwSize = ContainerSize(CONTAINER_ENUMERATION, twty, nItems);
    HGLOBAL hcon = GlobalAlloc(GMEM_MOVEABLE, dwSize);
    if (hcon) {
        pTW_ENUMERATION p = (pTW_ENUMERATION)GlobalLock(hcon);
        p->ItemType = (TW_UINT16)twty;
        p->NumItems = nItems;
        p->CurrentIndex = 0;
        p->DefaultIndex = 0;
        GlobalUnlock(hcon);
    }
    return WrapContainer(CONTAINER_ENUMERATION, hcon);
}


int EZTAPI CONTAINER_Format(HCONTAINER hcon)
// Return the 'format' of this container.
// (These are the same codes defined by the TWAIN standard)
{
    PEZCONTAINER pcon = (PEZCONTAINER)hcon;
    if (pcon == NULL || pcon->dwSig != CONTAINER_SIGNATURE) {
        return -1;
    }
    return pcon->nFormat;
}

int EZTAPI CONTAINER_ItemType(HCONTAINER hcon)
// Return the item type (what exact kind of data items are in the container.)
{
    PEZCONTAINER pcon = (PEZCONTAINER)hcon;
    if (!pcon || pcon->dwSig != CONTAINER_SIGNATURE || !pcon->pCon) {
        return -1;
    }
    //** Hack Alert **
    // All the containers put their ItemType field in the same
    // place, so it doesn't matter which union-member we use here.
    return pcon->pCon->onevalue.ItemType;
}

int EZTAPI CONTAINER_ItemCount(HCONTAINER hcon)
// Return the number of values in the container.
// For a ONEVALUE, return 1.
{
    PEZCONTAINER pcon = (PEZCONTAINER)hcon;
    if (pcon == NULL || pcon->dwSig != CONTAINER_SIGNATURE) {
        return -1;
    }
    return ItemCount(pcon);
} // CONTAINER_ItemCount


int EZTAPI CONTAINER_ContainsValue(HCONTAINER hcon, double d)
{
    return CONTAINER_FindValue(hcon, d) >= 0;
} // CONTAINER_ContainsValue


int EZTAPI CONTAINER_FindValue(HCONTAINER hcon, double d)
{
    PEZCONTAINER pcon = (PEZCONTAINER)hcon;
    int i;
    if (pcon == NULL || pcon->dwSig != CONTAINER_SIGNATURE) {
        return -1;
    }
    for (i = CONTAINER_ItemCount(hcon)-1;
         i >= 0 && CONTAINER_FloatValue(hcon, i) != d;
         i--) {}
    return i;
}

int EZTAPI CONTAINER_IntValue(HCONTAINER hcon, int n)
{
    return (int)EZround(CONTAINER_FloatValue(hcon, n));
    // Bugfix 3/14/00 - thanks to Dave Offen of Arkenstone.
    // C/C++ float-to-integral conversion rounds toward 0.
} // CONTAINER_IntValue


double EZTAPI CONTAINER_FloatValue(HCONTAINER hcon, int n)
{
    PEZCONTAINER pcon = (PEZCONTAINER)hcon;
    PTW_CONTAINER p;
    TW_UINT16 twty;
    int nValues;
    if (pcon == NULL || pcon->dwSig != CONTAINER_SIGNATURE) {
        return -1;
    }
    p = pcon->pCon;
    twty = p->onevalue.ItemType;
    // Clip n into valid range
    nValues = CONTAINER_ItemCount(hcon);
    n = min(nValues-1, max(0, n));
    switch (pcon->nFormat) {
    case CONTAINER_ONEVALUE:
        return Value2Float(p->onevalue.Item, twty);
    case CONTAINER_ENUMERATION:
        return Item2Float((BYTE*)&p->enumeration.ItemList + n * nTypeSize[twty], twty);
    case CONTAINER_ARRAY:
        return Item2Float((BYTE*)&p->array.ItemList + n * nTypeSize[twty], twty);
    case CONTAINER_RANGE:
        {
            double dMin = Value2Float(p->range.MinValue, twty);
            double dMax = Value2Float(p->range.MaxValue, twty);
            double dStep = Value2Float(p->range.StepSize, twty);
            double dVal = dMin + n * dStep;
            if (dVal > dMax) {
                dVal = dMax;
            }
            return dVal;
        }
    default:
        return -1.0;
    } // switch
} // CONTAINER_FloatValue


BYTE* EZTAPI CONTAINER_ValuePtr(HCONTAINER hcon, int n)
{
    PEZCONTAINER pcon = (PEZCONTAINER)hcon;
    PTW_CONTAINER p;
    TW_UINT16 twty;
    int nValues;
    if (pcon == NULL || pcon->dwSig != CONTAINER_SIGNATURE) {
        return NULL;
    }
    p = pcon->pCon;
    twty = p->onevalue.ItemType;
    // Clip n into valid range
    nValues = CONTAINER_ItemCount(hcon);
    n = min(nValues-1, max(0, n));
    switch (pcon->nFormat) {
    case CONTAINER_ONEVALUE:
        return (BYTE*)&p->onevalue.Item;
    case CONTAINER_ENUMERATION:
        return (BYTE*)&p->enumeration.ItemList + n * nTypeSize[twty];
    case CONTAINER_ARRAY:
        return (BYTE*)&p->array.ItemList + n * nTypeSize[twty];
    default:
        return NULL;
    } // switch
} // CONTAINER_ValuePtr


int EZTAPI CONTAINER_CurrentIndex(HCONTAINER hcon)
{
    PEZCONTAINER pcon = (PEZCONTAINER)hcon;
    if (pcon == NULL ||
        pcon->dwSig != CONTAINER_SIGNATURE ||
        pcon->nFormat != CONTAINER_ENUMERATION) {
        return -1;
    }
    return pcon->pCon->enumeration.CurrentIndex;
} // CONTAINER_CurrentIndex


int EZTAPI CONTAINER_DefaultIndex(HCONTAINER hcon)
{
    PEZCONTAINER pcon = (PEZCONTAINER)hcon;
    if (pcon == NULL ||
        pcon->dwSig != CONTAINER_SIGNATURE ||
        pcon->nFormat != CONTAINER_ENUMERATION) {
        return -1;
    }
    return pcon->pCon->enumeration.DefaultIndex;
} // CONTAINER_DefaultIndex


double EZTAPI CONTAINER_CurrentValue(HCONTAINER hcon)
{
    PEZCONTAINER pcon = (PEZCONTAINER)hcon;
    if (pcon != NULL && pcon->dwSig == CONTAINER_SIGNATURE) {
        PTW_CONTAINER p = pcon->pCon;
        TW_UINT16 twty = p->onevalue.ItemType;
        switch (pcon->nFormat) {
        case CONTAINER_ONEVALUE:
            return Value2Float(p->onevalue.Item, twty);

        case CONTAINER_ENUMERATION:
            if (p->enumeration.CurrentIndex < p->enumeration.NumItems) {
                return Item2Float((BYTE*)&p->enumeration.ItemList +
                                  p->enumeration.CurrentIndex * nTypeSize[twty], twty);
            }
            break;

        case CONTAINER_RANGE:
            return Value2Float(p->range.CurrentValue, twty);
            
        case CONTAINER_ARRAY:
        default:
            break;
        } // switch
    }
    return -1.0;
} // CONTAINER_CurrentValue


double EZTAPI CONTAINER_DefaultValue(HCONTAINER hcon)
{
    PEZCONTAINER pcon = (PEZCONTAINER)hcon;
    PTW_CONTAINER p;
    void* pv = NULL;
    TW_UINT16 twty;
    if (pcon == NULL ||
        pcon->dwSig != CONTAINER_SIGNATURE ||
        pcon->nFormat == CONTAINER_ARRAY) {
        return -1;
    }
    p = pcon->pCon;
    twty = p->onevalue.ItemType;
    switch (pcon->nFormat) {
    case CONTAINER_ONEVALUE:
        return Value2Float(p->onevalue.Item, twty);
    case CONTAINER_ENUMERATION:
        return Item2Float((BYTE*)&p->enumeration.ItemList +
                          p->enumeration.DefaultIndex * nTypeSize[twty], twty);
    case CONTAINER_RANGE:
        return Value2Float(p->range.DefaultValue, twty);
    default:
        return -1.0;
    } // switch
} // CONTAINER_DefaultValue


double EZTAPI CONTAINER_MinValue(HCONTAINER hcon)
{
    PEZCONTAINER pcon = (PEZCONTAINER)hcon;
    pTW_RANGE p;
    if (pcon == NULL ||
        pcon->dwSig != CONTAINER_SIGNATURE ||
        pcon->nFormat != CONTAINER_RANGE) {
        return -1;
    }
    p = (pTW_RANGE)pcon->pCon;
    return Value2Float(p->MinValue, p->ItemType);
} // CONTAINER_MinValue

double EZTAPI CONTAINER_MaxValue(HCONTAINER hcon)
{
    PEZCONTAINER pcon = (PEZCONTAINER)hcon;
    pTW_RANGE p;
    if (pcon == NULL ||
        pcon->dwSig != CONTAINER_SIGNATURE ||
        pcon->nFormat != CONTAINER_RANGE) {
        return -1;
    }
    p = (pTW_RANGE)pcon->pCon;
    return Value2Float(p->MaxValue, p->ItemType);
} // CONTAINER_MaxValue

double EZTAPI CONTAINER_StepSize(HCONTAINER hcon)
{
    PEZCONTAINER pcon = (PEZCONTAINER)hcon;
    pTW_RANGE p;
    if (pcon == NULL ||
        pcon->dwSig != CONTAINER_SIGNATURE ||
        pcon->nFormat != CONTAINER_RANGE) {
        return -1;
    }
    p = (pTW_RANGE)pcon->pCon;
    return Value2Float(p->StepSize, p->ItemType);
} // CONTAINER_StepSize


int EZTAPI CONTAINER_SelectDefaultValue(HCONTAINER hcon, double dVal)
{
    int n = CONTAINER_FindValue(hcon, dVal);
    return (n >= 0) && CONTAINER_SelectDefaultItem(hcon, n);
}


int EZTAPI CONTAINER_SelectDefaultItem(HCONTAINER hcon, int n)
{
    PEZCONTAINER pcon = (PEZCONTAINER)hcon;
    PTW_CONTAINER p;
    if (pcon == NULL ||
        pcon->dwSig != CONTAINER_SIGNATURE ||
        pcon->nFormat == CONTAINER_ONEVALUE ||
        pcon->nFormat == CONTAINER_ARRAY) {
        return FALSE;
    }
    if (n < 0 || n >= CONTAINER_ItemCount(hcon)) {
        return FALSE;
    }
    p = pcon->pCon;
    if (pcon->nFormat == CONTAINER_ENUMERATION) {
        p->enumeration.DefaultIndex = n;
    } else {
        // CONTAINER_RANGE:
        p->range.DefaultValue =
            Float2Value(CONTAINER_FloatValue(hcon, n),
                        p->range.ItemType);
    }
    return TRUE;
}


int EZTAPI CONTAINER_SelectCurrentValue(HCONTAINER hcon, double dVal)
{
    int n = CONTAINER_FindValue(hcon, dVal);
    return (n >= 0) && CONTAINER_SelectCurrentItem(hcon, n);
}


int EZTAPI CONTAINER_SelectCurrentItem(HCONTAINER hcon, int n)
{
    PEZCONTAINER pcon = (PEZCONTAINER)hcon;
    PTW_CONTAINER p;
    if (pcon == NULL ||
        pcon->dwSig != CONTAINER_SIGNATURE ||
        pcon->nFormat == CONTAINER_ONEVALUE ||
        pcon->nFormat == CONTAINER_ARRAY) {
        return FALSE;
    }
    if (n < 0 || n >= CONTAINER_ItemCount(hcon)) {
        return FALSE;
    }
    p = pcon->pCon;
    if (pcon->nFormat == CONTAINER_ENUMERATION) {
        p->enumeration.CurrentIndex = n;
    } else {
        // CONTAINER_RANGE:
        p->range.CurrentValue =
            Float2Value(CONTAINER_FloatValue(hcon, n),
                        p->range.ItemType);
    }
    return TRUE;
}


int EZTAPI CONTAINER_SetItem(HCONTAINER hcon, int n, double dV)
{
    PEZCONTAINER pcon = (PEZCONTAINER)hcon;
    PTW_CONTAINER p;
    TW_UINT16 twty;
    int nItems;
    if (pcon == NULL ||
        pcon->dwSig != CONTAINER_SIGNATURE ||
        pcon->nFormat == CONTAINER_RANGE) {
        return FALSE;
    }
    p = pcon->pCon;
    twty = p->onevalue.ItemType;
    // Clip n into valid range
    nItems = CONTAINER_ItemCount(hcon);
    if (n < 0 || n >= nItems) {
        return FALSE;
    }
    if (pcon->nFormat == CONTAINER_ENUMERATION) {
        return StoreItem((BYTE*)&p->enumeration.ItemList + n * nTypeSize[twty],
                         twty, dV);
    } else if (pcon->nFormat == CONTAINER_ONEVALUE) {
        return StoreItem((BYTE*)&p->onevalue.Item, twty, dV);
    } else {
        // CONTAINER_ARRAY
        return StoreItem((BYTE*)&p->array.ItemList + n * nTypeSize[twty],
                         twty, dV);
    }
} // CONTAINER_SetItem


int EZTAPI CONTAINER_SetItemString(HCONTAINER hcon, int n, LPCTSTR pzVal)
{
    PEZCONTAINER pcon = (PEZCONTAINER)hcon;
    PTW_CONTAINER p;
    BYTE* pitem;
    TW_UINT16 twty;
    int nItems;
    if (pcon == NULL ||
        pcon->dwSig != CONTAINER_SIGNATURE ||
        pcon->nFormat == CONTAINER_RANGE) {
        return FALSE;
    }
    p = pcon->pCon;
    twty = p->onevalue.ItemType;
    // Clip n into valid range
    nItems = CONTAINER_ItemCount(hcon);
    if (n < 0 || n >= nItems) {
        return FALSE;
    }
    if (pcon->nFormat == CONTAINER_ENUMERATION) {
        pitem = (BYTE*)&p->enumeration.ItemList + n * nTypeSize[twty];
    } else if (pcon->nFormat == CONTAINER_ONEVALUE) {
        pitem = (BYTE*)&p->onevalue.Item;
    } else {
        pitem = (BYTE*)&p->array.ItemList + n * nTypeSize[twty];
    }
#ifdef _UNICODE
    WideCharToMultiByte(
        CP_ACP,             // code page
        0,                  // flags
        pzVal,              // source wide char string
        -1,                 // length of same
        (LPSTR)pitem,       // destination ANSI string buffer
        nTypeSize[twty],    // size of buffer
        NULL,               // addr of default for unmappable chars
        NULL);              // addr of flag set when default used.
#else
    strncpy((LPSTR)pitem, pzVal, nTypeSize[twty]);
#endif
    return TRUE;
} // CONTAINER_SetItemString


int EZTAPI CONTAINER_SetItemFrame(HCONTAINER hcon, int n, double l, double t, double r, double b)
{
    PEZCONTAINER pcon = (PEZCONTAINER)hcon;
    PTW_CONTAINER p;
    TW_FRAME* pitem;
    TW_UINT16 twty;
    int nItems;
    if (pcon == NULL ||
        pcon->dwSig != CONTAINER_SIGNATURE ||
        pcon->nFormat == CONTAINER_RANGE) {
        return FALSE;
    }
    p = pcon->pCon;
    twty = p->onevalue.ItemType;
    if (twty != TWTY_FRAME) {
        return FALSE;
    }
    // Clip n into valid range
    nItems = CONTAINER_ItemCount(hcon);
    if (n < 0 || n >= nItems) {
        return FALSE;
    }
    if (pcon->nFormat == CONTAINER_ENUMERATION) {
        pitem = (TW_FRAME*)&p->enumeration.ItemList + n * nTypeSize[twty];
    } else if (pcon->nFormat == CONTAINER_ONEVALUE) {
        pitem = (TW_FRAME*)&p->onevalue.Item;
    } else {
        pitem = (TW_FRAME*)&p->array.ItemList + n * nTypeSize[twty];
    }
    SetFix32(&pitem->Left, l);
    SetFix32(&pitem->Top, t);
    SetFix32(&pitem->Right, r);
    SetFix32(&pitem->Bottom, b);
    return TRUE;
} // CONTAINER_SetItemFrame


int EZTAPI CONTAINER_DeleteItem(HCONTAINER hcon, int n)
// Delete the nth item from an Array or Enumeration container.
// Returns 1 (TRUE) for success, 0 (FALSE) otherwise. Failure causes:
//   invalid container handle
//   container is not an array or enumeration
//   n < 0 or n >= ItemCount(hcon)
{
    PEZCONTAINER pcon = (PEZCONTAINER)hcon;
    PTW_CONTAINER p;
    BYTE* pitem;
    TW_UINT16 twty;
    DWORD cbItem;
    int nItems;
    if (pcon == NULL ||
        pcon->dwSig != CONTAINER_SIGNATURE ||
        pcon->nFormat == CONTAINER_ONEVALUE ||
        pcon->nFormat == CONTAINER_RANGE) {
        return FALSE;
    }
    p = pcon->pCon;
    twty = p->onevalue.ItemType;
    cbItem = nTypeSize[twty];
    // Clip n into valid range
    nItems = CONTAINER_ItemCount(hcon);
    if (n < 0 || n >= nItems) {
        return FALSE;
    }
    if (pcon->nFormat == CONTAINER_ENUMERATION) {
        pitem = (BYTE*)&p->enumeration.ItemList + n * cbItem;
    } else {
        // CONTAINER_ARRAY
        pitem = (BYTE*)&p->array.ItemList + n * cbItem;
    }
    FMEMCPY(pitem, pitem+cbItem, (nItems - 1 - n) * cbItem);
    return TRUE;
} // CONTAINER_DeleteItem


int EZTAPI CONTAINER_InsertItem(HCONTAINER hcon, int n, double dVal)
// Insert an item with value dVal into the container at position n.
// If n < 0, the item is inserted at the end of the container.
// Return 1 (TRUE) on success, 0 (FALSE) on failure.  Failure causes:
//   NULL or invalid container handle
//   container format is not enumeration or array
//   memory allocation failed (not likely under Win 9x/NT)
{
    PEZCONTAINER pcon = (PEZCONTAINER)hcon;
    PTW_CONTAINER p;
    BYTE* pitem;
    TW_UINT16 twty;
    DWORD cbItem;
    DWORD dwOldSize, dwNewSize;
    int nItems;

    if (pcon == NULL ||
        pcon->dwSig != CONTAINER_SIGNATURE ||
        pcon->nFormat == CONTAINER_ONEVALUE ||
        pcon->nFormat == CONTAINER_RANGE) {
        return FALSE;
    }
    p = pcon->pCon;
    twty = p->onevalue.ItemType;
    cbItem = nTypeSize[twty];
    // Clip n into valid range
    nItems = ItemCount(pcon);
    if (n < 0 || n >= nItems) {
        n = nItems;
    }
    dwOldSize = GlobalSize(pcon->hCon);
    dwNewSize = ContainerSize(pcon->nFormat, twty, nItems+1);
    if (dwNewSize > dwOldSize) {
        HGLOBAL hnew;
        GlobalUnlock(pcon->hCon);
        hnew = GlobalReAlloc(pcon->hCon, dwNewSize+128, GMEM_MOVEABLE);
        if (hnew == NULL) {
            pcon->pCon = GlobalLock(pcon->hCon);
            return FALSE;
        }
        pcon->hCon = hnew;
        p = pcon->pCon = GlobalLock(pcon->hCon);
    }
    if (pcon->nFormat == CONTAINER_ENUMERATION) {
        pitem = (BYTE*)&p->enumeration.ItemList + n * cbItem;
        p->enumeration.NumItems++;
    } else {
        // CONTAINER_ARRAY
        pitem = (BYTE*)&p->array.ItemList + n * cbItem;
        p->array.NumItems++;
    }
    FMEMCPY(pitem+cbItem, pitem, (nItems - n) * cbItem);
    return StoreItem(pitem, twty, dVal);
} // CONTAINER_InsertItem


/////////////////////////////////////////////////////////////////////////////
// Container SET/GET Functions
/////////////////////////////////////////////////////////////////////////////

HCONTAINER EZTAPI MsgGet(unsigned uCap, unsigned uMsg)
{
	TW_CAPABILITY 	cap;
	if (TWAIN_State() < TWAIN_SOURCE_OPEN) {
		TWAIN_ErrorBox("TWAIN_GetXXX below State 4.");
		return 0;
	}

	// Fill in capability structure
	cap.Cap = (TW_UINT16)uCap;					// capability id
	cap.ConType = TWON_DONTCARE16;  // up to DS
	cap.hContainer = NULL;

    if (!TWAIN_DS(DG_CONTROL, DAT_CAPABILITY, uMsg, (TW_MEMREF)&cap) ||
        cap.hContainer == NULL) {
        // GET failed or container handle is NULL
        return 0;
    }
    return WrapContainer(cap.ConType, cap.hContainer);
} // MsgGet


HCONTAINER EZTAPI TWAIN_Get(unsigned uCap)
{
    return MsgGet(uCap, MSG_GET);
} // TWAIN_Get


HCONTAINER EZTAPI TWAIN_GetDefault(unsigned uCap)
{
    return MsgGet(uCap, MSG_GETDEFAULT);
} // TWAIN_GetDefault


HCONTAINER EZTAPI TWAIN_GetCurrent(unsigned uCap)
{
    return MsgGet(uCap, MSG_GETCURRENT);
} // TWAIN_GetCurrent


int EZTAPI TWAIN_Set(unsigned uCap, HCONTAINER hcon)
{
    TW_CAPABILITY 	cap;
    PEZCONTAINER pcon = (PEZCONTAINER)hcon;
    if (pcon == NULL || pcon->dwSig != CONTAINER_SIGNATURE) {
        return FALSE;
    }
	if (TWAIN_State() < TWAIN_SOURCE_OPEN) {
		TWAIN_ErrorBox("Setting capability in State < 4.");
		return FALSE;
	}

	// Fill in a capability structure
	cap.Cap = (TW_UINT16)uCap;					// capability id
    cap.ConType = (TW_UINT16)(pcon->nFormat);   // container format
	cap.hContainer = pcon->hCon;    // container handle

    return TWAIN_DS(DG_CONTROL, DAT_CAPABILITY, MSG_SET, (TW_MEMREF)&cap);
} // TWAIN_Set


