///////////////////////////////////////////////////////////////////////
// EZTWAIN.C - EZTWAIN DLL, main module
// Copyright (C) 1999-2001 by Dosadi.  All Rights Reserved.
//
// Download from:    www.dosadi.com
//
// Support contact:  support@dosadi.com
//
// See end of this file for revision history
///////////////////////////////////////////////////////////////////////


#define VERSION 270		   // version number, times 100.
#ifdef _DEBUG
#define BUILDNAME "Debug"
#else
#define BUILDNAME "Release"   // build name
#endif

#define IJL_JPEG // Intel JPEG Library support

// fussy checking (and whining) during memory transfers
#ifdef _DEBUG
#define MEM_FUSSY
#define LOG_FILE
#endif

///////////////////////////////////////////////////////////////////////
//------------ Includes

#include <assert.h>
#include <memory.h>		// memset, memcpy
#include <stdio.h>      // sprintf
#include <io.h>         // for _lseek
#include <crtdbg.h>     // _CrtIsValidPointer
#include <time.h>
#include "windows.h"

#ifdef _WIN32
#ifndef _INC_VFW
#include <vfw.h>
#endif
#else
#include "commdlg.h"
#endif

#include "twain.h"

#ifdef _WIN32
  #define EZTAPI WINAPI
#else
  #define EZTAPI __export FAR PASCAL
#endif

#include "eztwain.h"

#ifdef IJL_JPEG
#include "ijl.h"
#endif

///////////////////////////////////////////////////////////////////////
// Typedefs

typedef RGBQUAD FAR *LPQUAD;

///////////////////////////////////////////////////////////////////////
//------------ Shared with other EZTWAIN modules

extern long EZround(double d);
extern double Fix32ToFloat(TW_FIX32 fix);
extern void SetFix32(TW_FIX32* pfix, double r);

FILE* pfsLog;             // tracing log


///////////////////////////////////////////////////////////////////////
// Imports from other EZTWAIN modules

extern double PPM2DPI(long ppm);
extern long DPI2PPM(double dpi);

int CopyMemoryBufferToDib(HANDLE hdib,
                          TW_IMAGEMEMXFER mx,
                          TW_IMAGEINFO info,
                          int nChannel);
int BmiWriteToBmpFile(LPBITMAPINFOHEADER lpDIB, HFILE fh);
LPBYTE BmiFindBits(LPBITMAPINFOHEADER pbi);
long BmiRowBytes(const LPBITMAPINFOHEADER pbi);
int BmiHeight(const LPBITMAPINFOHEADER pbi);


///////////////////////////////////////////////////////////////////////
//------------ Forward declarations

void EZTAPI TWAIN_WaitForXfer(HWND hwnd);
void EZTAPI TWAIN_EmptyMessageQueue(void);
void EZTAPI TWAIN_BreakModalLoop(void);

static void SetState(int n);
static HWND GetValidHwnd(HWND hwnd);
static HWND CreateProxyWindow(void);
static TW_UINT32 Intersect(TW_UINT32 wMask, TW_UINT32 nItems, TW_UINT16 far *pItem);
static unsigned FindIndex16(TW_UINT32 nItems, TW_UINT16 far *plist, TW_UINT16 uVal, unsigned nDefault);
static unsigned BitCount(DWORD W);
static int RecordError(EZTWAIN_ERROR_CODE ed);
static void NativeXferHandler(void);
static void FileXferHandler(void);
BOOL WINAPI MemoryXferHandler(LPVOID pdata);
static void DoOneTransfer(void);
static BOOL PromptForFilename(char* pszFile, int* pnFF);
#ifdef IJL_JPEG
static int LoadJpegSupport(void);
#endif

#ifdef EZTPRO_H
static HWND FindScanButton(void);
static void PressButton(HWND hwnd);
#endif

HANDLE EZTAPI TWAIN_LoadNativeFromFile(HFILE fh);

///////////////////////////////////////////////////////////////////////
//------------ Constants and Macros

#define STATIC static

#ifdef _WIN32
  #define INT32 int
  #define DATAGROUP unsigned
  #define DSM_FILENAME "TWAIN_32.DLL"
  #define DSM_ENTRYPOINT "DSM_Entry"
  #define IsValidHandle(h) (h!=NULL)
  #define HUGEWRITE(fh, pb, bc) ((INT32)_lwrite(fh, pb, bc))
  #define HUGEREAD(fh, pb, bc) ((INT32)_lread(fh, pb, bc))
  #define FMEMSET(p, v, n) memset(p, v, n);
  #define FMEMCPY(p, q, n) memcpy(p, q, n);
#else
  #define INT32 long
  #define DATAGROUP unsigned long
  #define DSM_FILENAME "TWAIN.DLL"
  #define DSM_ENTRYPOINT "DSM_ENTRY"
  #define VALID_HANDLE 32
  #define IsValidHandle(h) ((h!=0) && ((h) >= 32))
  #define HUGEWRITE(fh, pb, bc) _hwrite(fh, pb, bc)
  #define HUGEREAD(fh, pb, bc) _hread(fh, pb, bc)
  #define FMEMSET(p, v, n) _fmemset(p, v, n);
  #define FMEMCPY(p, q, n) _fmemcpy(p, q, n);
#endif

const size_t nTypeSize[13] =
	{	sizeof (TW_INT8),
		sizeof (TW_INT16),
		sizeof (TW_INT32),
		sizeof (TW_UINT8),
		sizeof (TW_UINT16),
		sizeof (TW_UINT32),
		sizeof (TW_BOOL),
		sizeof (TW_FIX32),
		sizeof (TW_FRAME),
		sizeof (TW_STR32),
		sizeof (TW_STR64),
		sizeof (TW_STR128),
		sizeof (TW_STR255),
	};

const char *pszErrDescrip[] = 
	{	"[no details available]",
		"",
		"DAT_CAPABILITY/MSG_GET failed",
		"DAT_CAPABILITY/MSG_SET failed",
		"TWAIN Source Manager operation failed",
		"DataSource operation failed",
        "Image Transfer failed",
		"",
        "Opening TWAIN Source Manager failed",
        "Unable to open default TWAIN DataSource",
		"TWAIN session not in State 4 (Source Open)",
		"MSG_GET returned a NULL container handle",
		"MSG_GET returned an invalid container handle",
		"Returned container is not valid type",
		"Returned container has invalid ItemType",
		"Returned container is empty",
		"App and source found NO values in common",
        "EZTWAIN function called with invalid window handle",
        "EZTWAIN proxy window create failed",
        "Operation cancelled by user",
        "Device rejected setting for Resolution"
	};

const char *pszRC[] = {
    "TWRC_SUCCESS",
    "TWRC_FAILURE",
    "TWRC_CHECKSTATUS ('tried hard')",
    "TWRC_CANCEL",
    "TWRC_DSEVENT",
    "TWRC_NOTDSEVENT",
    "TWRC_XFERDONE",
    "TWRC_ENDOFLIST",
    "TWRC_INFONOTSUPPORTED",
    "TWRC_DATANOTAVAILABLE"
};

const char *pszCC[] = {
	"TWCC_SUCCESS",
	"TWCC_BUMMER (Failure due to unknown causes)",
	"TWCC_LOWMEMORY",
	"TWCC_NODS (No Data Source)",
	"TWCC_MAXCONNECTIONS (DS is connected to max possible apps)",
	"TWCC_OPERATIONERROR (DS/DSM reported error, app shouldn't)",
	"TWCC_BADCAP (Unknown capability)",
	"7 (undefined)",
	"8 (undefined)",
	"TWCC_BADPROTOCOL (Unrecognized triplet)",
	"TWCC_BADVALUE (Data parameter out of range)",
	"TWCC_SEQERROR (Triplet out of sequence)",
	"TWCC_BADDEST (Unknown dest. App/Src in DSM_Entry)",
   "TWCC_CAPUNSUPPORTED (Cap not supported by source)",
   "TWCC_CAPBADOPERATION (Operation not supported by cap)",
   "TWCC_CAPSEQERROR (Cap has dependancy on other cap)",
};

const char szInsuffMem[] = "Insufficient Memory";	// error message
const char szBadType[] = "TWAIN_SetCapOneValue: invalid item type.";

static char szBuildName[] = BUILDNAME;

#define WM_AUTOSCAN 0x7fa0       // semi-random number

///////////////////////////////////////////////////////////////////////
//------------ Global variables

STATIC int          nAttached;          // number of Attaches - number of Detaches
STATIC int			  iAvailable;			 // TWAIN available: 0:unknown, -1:No, 1:Yes
STATIC int			  nState = 1;			 // TWAIN state (per the standard)
STATIC int			  nErrDetail;			 // detailed error code
STATIC unsigned     nErrRC, nErrCC;	    // result code and condition code for last error
STATIC BOOL         bCloseDsRequested;
STATIC char         szMsg[256];			 // scratch buffer for messages
STATIC DSMENTRYPROC pDSM_Entry;         // entry point of Data Source Manager (TWAIN.DLL)
STATIC HANDLE       hDSMLib;            // handle of DSM
STATIC TW_IDENTITY  AppId = {           // application identity structure
	0,									// Id, filled in by DSM
	{ 1, 0, TWLG_USA, TWCY_USA, "Unknown"},	// Version
    TWON_PROTOCOLMAJOR,
    TWON_PROTOCOLMINOR,
    DG_IMAGE | DG_CONTROL,
    "Unknown",                            // Mfg
    "Unspecified",                        // Family
    "TWAIN Application"                   // Product
    };
STATIC TW_IDENTITY      SourceId;         // source identity structure
STATIC BOOL             bHideUI;          // allow source u/i to be hidden
STATIC BOOL             bIndicators = TRUE; // show (hide) progress indicators
STATIC BOOL             bDuplex = -1;       // enable/disable Duplex scanning
STATIC BOOL             bAutoFeed = -1;     // enable/disable CAP_AUTOFEED
STATIC double           dxResolution = -1.0;
STATIC double           dyResolution = -1.0;

STATIC TW_USERINTERFACE twUI;
STATIC TW_PENDINGXFERS  pendingXfers;
STATIC BOOL             bBreakModalLoop;  // flag to end modal event loop
STATIC int              nXferMech = XFERMECH_NATIVE;
STATIC unsigned         nMemBuffer;       // memory transfer buffer counter
STATIC HANDLE           hDib;             // bitmap returned by native transfer
STATIC TW_INT16         rc;               // result code       
STATIC HINSTANCE        hinstLib;         // instance handle for this DLL
STATIC HWND             hwndProxy;        // proxy window
STATIC BOOL             bMultiTransfer;   // allow multi-image transfer per session
STATIC BOOL             bNoMessages;      // suppress EZTWAIN error messages
STATIC int              nSaveFormat = TWFF_BMP;    // output file format
STATIC int              nFileXferFormat;
STATIC char             szFileXferPath[_MAX_PATH];
STATIC BOOL             bXferDone;

STATIC MEMXFERCALLBACK  cbXFunc = MemoryXferHandler;
STATIC LPVOID			cbData;
STATIC BOOL				bCBSuccess;		
STATIC BOOL				bDisableParent = TRUE;

#ifdef IJL_JPEG
// Intel JPEG Library globals
typedef IJLERR (WINAPI *PIJLINIT)( JPEG_CORE_PROPERTIES*);
typedef IJLERR (WINAPI *PIJLIO)( JPEG_CORE_PROPERTIES*, IJLIOTYPE);
static HMODULE hijl;
static PIJLINIT pijlInit, pijlFree;
static PIJLIO pijlWrite, pijlRead;
#endif

#ifdef EZTPRO_H
STATIC HWND             hwndButton;
STATIC BOOL             bAutoScan;
#endif

///////////////////////////////////////////////////////////////////////
//------------ Public functions

int EZTWAIN_Attach(void)
{
   if (0 == nAttached++) {
      hinstLib = GetModuleHandle(NULL);
      AppId.Id = 0;
      FMEMSET(&SourceId, 0, sizeof SourceId);
#ifdef _DEBUG
      wsprintf(szMsg, "sizeof (TW_IDENTITY) == %d\n", sizeof (TW_IDENTITY));
      OutputDebugString(szMsg);
#endif
      assert(sizeof (TW_IDENTITY) == 156);
#ifdef LOG_FILE
      TWAIN_LogFile(1);
#endif
   }
   return nAttached;
} // EZTWAIN_Attach


int EZTWAIN_Detach(void)
{
   if (nAttached == 1) {
      SetState(TWAIN_PRESESSION);
#ifdef LOG_FILE
      TWAIN_LogFile(0);
#endif
      if (hDSMLib) {
         FreeLibrary(hDSMLib);
         hDSMLib = NULL;
      }
      pDSM_Entry = NULL;
#ifdef IJL_JPEG
      if (hijl) FreeLibrary(hijl);
#endif
   }
   nAttached--;
   return nAttached;
} // EZTWAIN_Detach


#ifdef _WIN32

// Win32 DLL main
BOOL WINAPI DllMain(HANDLE hModule, 
                    ULONG  ulEvent,
                    LPVOID lpReserved)
{
    lpReserved = lpReserved;        // suppress unref param warning
    
    switch (ulEvent) {

    case DLL_PROCESS_ATTACH:
        EZTWAIN_Attach();
        hinstLib = hModule;
        break;

    case DLL_PROCESS_DETACH:
        EZTWAIN_Detach();
        break;

    default:
        break;
    }
    return TRUE;
}

#else

// Win16 DLL initialization and termination routines
int CALLBACK __export LibMain(HINSTANCE hinst, WORD wDataSeg, WORD cbHeapSize, LPSTR lpszCmdLine)
{
    wDataSeg = wDataSeg; cbHeapSize = cbHeapSize; lpszCmdLine = lpszCmdLine;
    
    hinstLib = hinst;
    assert(sizeof (TW_IDENTITY) == 156);
    nAttached++;
    return 1;		// indicate success
} // LibMain


int FAR PASCAL __export _WEP(int x)
{
	x = x;				// suppress 'is never used' warning
    nAttached--;
	return 1;
} // _WEP


#endif


//-- Test calling interface
HANDLE EZTAPI TWAIN_Testing123(LPCSTR pz, int n, HANDLE h, double d, unsigned u)
{
   sprintf(szMsg, "pz=\"%hs\"\nn=%d\nh=%08lx (hex)\nd=%g\nu=%u",
      pz, n, (unsigned long)h, d, u);
	MessageBox(NULL, szMsg, "TWAIN_Testing123", MB_ICONINFORMATION | MB_OK);
   return h;
} // TWAIN_Testing123

void EZTAPI TWAIN_SetAppTitle(LPCSTR lpszTitle)
{
    strncpy(AppId.ProductName, lpszTitle, 32);
} // TWAIN_SetAppTitle


//-- Record application information
void EZTAPI TWAIN_RegisterApp(
	int   nMajorNum,     // major and incremental revision of application.
    int   nMinorNum,     // e.g. version 4.5: nMajorNum = 4, nMinorNum = 5
	int   nLanguage,     // (human) language (use TWLG_xxx from TWAIN.H)
	int   nCountry,      // country (use TWCY_xxx from TWAIN.H)
	LPCSTR lpszVersion,   // version info string e.g. "1.0b3 Beta release"
	LPCSTR lpszMfg,       // name of mfg/developer e.g. "Crazbat Software"
	LPCSTR lpszFamily,    // product family e.g. "BitStomper"
	LPCSTR lpszProduct)   // specific product e.g. "BitStomper Deluxe Pro"
{
    AppId.Id = 0;						// init to 0, but Source Manager will assign real value
    AppId.Version.MajorNum = (TW_UINT16)nMajorNum;
    AppId.Version.MinorNum = (TW_UINT16)nMinorNum;
    AppId.Version.Language = (TW_UINT16)nLanguage;
    AppId.Version.Country  = (TW_UINT16)nCountry;
    strncpy(AppId.Version.Info,  lpszVersion, 32);

    AppId.ProtocolMajor =    TWON_PROTOCOLMAJOR;
    AppId.ProtocolMinor =    TWON_PROTOCOLMINOR;
    AppId.SupportedGroups =  DG_IMAGE | DG_CONTROL;
    strncpy(AppId.Manufacturer,  lpszMfg, 32);
    strncpy(AppId.ProductFamily, lpszFamily, 32);
    strncpy(AppId.ProductName,   lpszProduct, 32);

} // TWAIN_RegisterApp


int EZTAPI TWAIN_SelectImageSource(HWND hwnd)
{
   int fSuccess = FALSE;
   int nEntryState = nState;
   
   if (nState >= TWAIN_SM_OPEN || TWAIN_OpenSourceManager(hwnd)) {
      TW_IDENTITY	NewSourceId;
#ifdef _WIN32
      FMEMSET(&NewSourceId, 0, sizeof NewSourceId);
#else
      // I will settle for the system default.  Shouldn't I get a highlight
      // on system default without this call?
      TWAIN_Mgr(DG_CONTROL, DAT_IDENTITY, MSG_GETDEFAULT, &NewSourceId);
#endif
      // Post the Select Source dialog
      fSuccess = TWAIN_Mgr(DG_CONTROL, DAT_IDENTITY, MSG_USERSELECT, &NewSourceId);
   } else {
      TWAIN_ErrorBox("Unable to open Source Manager (" DSM_FILENAME ")");
   }
   
   if (nEntryState < TWAIN_SM_OPEN) {
      TWAIN_CloseSourceManager(hwnd);
      if (nEntryState < TWAIN_SM_LOADED) {
         TWAIN_UnloadSourceManager();
      }
   }
   return fSuccess;
} // TWAIN_SelectImageSource


static BOOL PreAcquire(HWND hwnd, int nXferMech, unsigned wPixTypes)
{
    TWAIN_ClearError();			// clear error detail
    bCloseDsRequested = FALSE;

    if (nState < TWAIN_SOURCE_OPEN) {
        // Bring up to state 4
        if (!TWAIN_OpenSourceManager(hwnd)) {
            TWAIN_ErrorBox("Unable to open Source Manager (" DSM_FILENAME ")");
            return FALSE;
        }
        if (!TWAIN_OpenDefaultSource()) {
            TWAIN_ReportLastError("Unable to open default Data Source.");
            return FALSE;
        }
        assert(nState == TWAIN_SOURCE_OPEN);
    }
    if (nState == TWAIN_SOURCE_OPEN) {
        // Negotiate transfer mode
        if (!TWAIN_SetXferMech(nXferMech)) {
            TWAIN_ReportLastError("Error setting transfer mode");
            return FALSE;
        }
        if (bDuplex >= 0) {
            // Try to negotiate CAP_DUPLEXENABLED
            if (!TWAIN_EnableDuplex(bDuplex)) {
                TWAIN_ReportLastError("Failed to negotiate Duplex.");
                return FALSE;
            }
        }
        if (bAutoFeed >= 0) {
            // Try to negotiate CAP_AUTOFEED
            if (!TWAIN_SetAutoFeed(bAutoFeed)) {
                TWAIN_ReportLastError("Failed to negotiate AutoFeed.");
                return FALSE;
            }
        }
        if (!TWAIN_NegotiatePixelTypes(wPixTypes)) {
            TWAIN_ReportLastError("Failed to negotiate Pixel Type.");
            return FALSE;
        }
        // Try to negotiate resolution, if requested
        if ((dxResolution >= 0.0 && !TWAIN_SetXResolution(dxResolution)) ||
            (dyResolution >= 0.0 && !TWAIN_SetYResolution(dyResolution))) {
            RecordError(EZTEC_RESOLUTION);
            TWAIN_ReportLastError("Failed to negotiate Resolution.");
            return FALSE;
        }

        // Try to negotiate CAP_INDICATORS:
        TWAIN_SetIndicators(bIndicators);
        // For memory transfers, request uncompressed data:
        if (nXferMech == XFERMECH_MEMORY) {
            // Try to avoid a transfer we can't handle
            // We can't handle compressed data
            TWAIN_SetCompression(TWCP_NONE);
        }
    }
    return (nState >= TWAIN_SOURCE_OPEN);
} // PreAcquire


static void PostAcquire(void)
{
    if (!bMultiTransfer) {
        // shut everything down in the right sequence
        TWAIN_UnloadSourceManager();
    } else {
        TWAIN_EmptyMessageQueue();
    }
} // PostAcquire


BOOL EZTAPI TWAIN_AcquireMemoryCallback(HWND hwnd, MEMXFERCALLBACK cb, LPVOID data)
{
    bCBSuccess = FALSE;
    if (cb == NULL || IsBadCodePtr(cb)) {
        TWAIN_ErrorBox("AcquireCB: Invalid callback address");
        return FALSE;
    }
    cbXFunc = cb;
    cbData = data;
    hwnd = GetValidHwnd(hwnd);
    if (PreAcquire(hwnd, XFERMECH_MEMORY, 0)) {
        TWAIN_WaitForNativeXfer(hwnd);
    }
    PostAcquire();
    // Restore default transfer handler:
    cbXFunc = MemoryXferHandler;
    // Note: bRetval will be FALSE if transfer failed or was cancelled.
    return bCBSuccess;
} // TWAIN_AcquireMemoryCallback

	
HANDLE EZTAPI TWAIN_Acquire(HWND hwnd)
{
    HANDLE hnative = NULL;
    hwnd = GetValidHwnd(hwnd);
    if (PreAcquire(hwnd, XFERMECH_NATIVE, 0)) {
        hnative = TWAIN_WaitForNativeXfer(hwnd);
    }
    PostAcquire();
    // Note: hnative will be NULL if transfer failed or was cancelled.
    return hnative;
} // TWAIN_Acquire


int EZTAPI TWAIN_AcquireFile(HWND hwnd, int nFF, LPCSTR pszFile)
{
    hwnd = GetValidHwnd(hwnd);
    nFileXferFormat = nFF;
    GetFullPathName(pszFile, _MAX_PATH, szFileXferPath, NULL);
    bXferDone = FALSE;
    if (PreAcquire(hwnd, XFERMECH_FILE, 0)) {
        TWAIN_WaitForXfer(hwnd);
    }
    PostAcquire();
    return bXferDone;
} // TWAIN_AcquireFile



HANDLE EZTAPI TWAIN_AcquireNative(HWND hwnd, unsigned wPixTypes)
{
    HANDLE hnative = NULL;
    hwnd = GetValidHwnd(hwnd);
    if (PreAcquire(hwnd, XFERMECH_NATIVE, wPixTypes)) {
        hnative = TWAIN_WaitForNativeXfer(hwnd);
    }
    PostAcquire();    
    // Note: hnative will be NULL if transfer failed or was cancelled.
    return hnative;
} // TWAIN_AcquireNative


HANDLE EZTAPI TWAIN_AcquireMemory(HWND hwnd)
{
    HANDLE hnative = NULL;
    hwnd = GetValidHwnd(hwnd);
    if (PreAcquire(hwnd, XFERMECH_MEMORY, 0)) {
        hnative = TWAIN_WaitForNativeXfer(hwnd);
    }
    PostAcquire();
    // Note: hnative will be NULL if transfer failed or was cancelled.
    return hnative;
} // TWAIN_AcquireMemory



int EZTAPI TWAIN_IsTransferReady(void)
{
    return (hDib != NULL) || (nState >= TWAIN_TRANSFER_READY);
} // IsTransferReady


HANDLE EZTAPI TWAIN_WaitForNativeXfer(HWND hwnd)
{
    HANDLE handle;
    if (!hDib) {
        if (nState >= TWAIN_SOURCE_OPEN) {
            TWAIN_WaitForXfer(hwnd);
        } else {
            TWAIN_ErrorBox("TWAIN_WaitForNativeXfer called in state < 4.");
        }
    }
    handle = hDib; hDib = NULL;
    return handle;
} // TWAIN_WaitForNativeXfer


void EZTAPI TWAIN_WaitForXfer(HWND hwnd)
{
   BOOL bWasEnabled;
   // Make up a valid window if we weren't given one
   hwnd = GetValidHwnd(hwnd);
   if (bDisableParent) {
       // Disable the parent window during the modal acquire
       bWasEnabled = (EnableWindow(hwnd, FALSE) == 0);
   }

   if (nState == TWAIN_TRANSFER_READY) {
      DoOneTransfer();
   } else if (nState >= TWAIN_SOURCE_ENABLED  || TWAIN_EnableSource(hwnd)) {
      // source is enabled, wait for transfer or source closed
      TWAIN_ModalEventLoop();
   } else {
      TWAIN_ReportLastError("Failed to enable Data Source.");
   }

   if (bDisableParent) {
       // Re-enable the parent window if it was enabled
       EnableWindow(hwnd, bWasEnabled);
   }
} // TWAIN_WaitForXfer


void EZTAPI TWAIN_FreeNative(HANDLE hdib)
// Release the memory allocated to a native format image, as returned by TWAIN_AcquireNative.
// (For those coding in C or C++, this is just a call to GlobalFree.)
{
	if (hdib) GlobalFree(hdib);
} // TWAIN_FreeNative


int EZTAPI TWAIN_AcquireToClipboard(HWND hwndApp, unsigned wPixTypes)
// Like AcquireNative, but puts the resulting image, if any, into the system clipboard.
// Useful for environments like Visual Basic where it is hard to make direct use of a DIB handle.
// A return value of 1 indicates success, 0 indicates failure.
{
	int nRes = 0;
    HANDLE hdib;
    if (nState >= TWAIN_SOURCE_OPEN || TWAIN_OpenDefaultSource()) {
        if (nState == TWAIN_SOURCE_OPEN) {
            TWAIN_NegotiatePixelTypes(wPixTypes);
        }
        hdib = TWAIN_Acquire(hwndApp);
        if (hdib) {
            nRes = DIB_PutOnClipboard(hdib);
            if (!nRes) {
                // something went wrong, recycle the image
                GlobalFree(hdib);
            }
        }
    }
	return nRes;
} // TWAIN_AcquireToClipboard


int EZTAPI TWAIN_AcquireToFilename(HWND hwndApp, LPCSTR pszFile)
// Adapted from a routine by David D. Henseler (ddh) of SOLUTIONS GmbH
{
	int result = -1;
	HANDLE	hnative = TWAIN_AcquireNative(hwndApp, 0);
	if (hnative) {
		result = TWAIN_WriteNativeToFilename(hnative, pszFile);
        if (result == -1) {
            result = -10;
        }
		TWAIN_FreeNative(hnative);
	}
	return result;
} // TWAIN_AcquireToFilename


///////////////////////////////////////////////////////////////////////
// TWAIN general


int EZTAPI TWAIN_IsAvailable(void)
// return 1 if TWAIN services are available, 0 if 'TWAIN-less' system
{
	if (pDSM_Entry) return TRUE;		// SM currently loaded

	if (iAvailable == 0) {
		if (TWAIN_LoadSourceManager()) {
			iAvailable = 1;
			TWAIN_UnloadSourceManager();
		} else {
			iAvailable = -1;
		}
	}
	return (iAvailable > 0);
} // TWAIN_IsAvailable


int EZTAPI TWAIN_EasyVersion(void)
// Returns the version number of EZTWAIN.DLL, multiplied by 100.
// So e.g. version 2.01 will return 201 from this call.
{
	return VERSION;
} // TWAIN_EasyVersion


char* EZTAPI TWAIN_BuildName(void)
{
    return szBuildName;
} // TWAIN_BuildName

void EZTAPI TWAIN_GetBuildName(LPSTR psName)
{
    strcpy(psName, szBuildName);
}

int EZTAPI TWAIN_SelfTest(unsigned f)
{
    f = f;              // suppress unref param warning

    if (sizeof (TW_IDENTITY) != 156) {
        return -1;
    }
    if (sizeof (char) != 1) {
        return -2;
    }
    if (sizeof (int) != 2 && sizeof (int) != 4) {
        return -3;
    }
    if (sizeof (long) != 4) {
        return -4;
    }
    if (sizeof (DWORD) != 4) {
        return -5;
    }
    if (TWAIN_EasyVersion() != VERSION) {
        return -101;
    }
    if (iAvailable > 1 || iAvailable < -1) {
        return -102;
    }
    if (TWAIN_State() < 1 || TWAIN_State() > 7) {
        return -103;
    }
    // Check rounding (new function as of today...)
    if (EZround(0.0) != 0.0 ||
        EZround(0.1) != 0.0 ||
        EZround(0.4999) != 0.0 ||
        EZround(0.5) != 1.0 ||
        EZround(-0.0001) != 0.0 ||
        EZround(-0.4999) != 0.0 ||
        EZround(-0.5) != -1.0) {
        return -104;
    }
    {
        // Check the basic integer fix-to-float stuff
        double v, v2;
        for (v = -32000; v <= 32000; v += 0.125) {
            long lx = TWAIN_ToFix32(v);
            v2 = TWAIN_Fix32ToFloat(lx);
            if (v2 != v) {
                return -105;
            }
        }
    }
    return 0;
}

int EZTAPI TWAIN_State(void)
// Returns the TWAIN Protocol State per the spec.
{
	return nState;
} 

void EZTAPI TWAIN_SetMultiTransfer(int fYes)
{
    bMultiTransfer = (fYes != 0);
} // TWAIN_SetMultiTransfer

int EZTAPI TWAIN_GetMultiTransfer(void)
{
    return bMultiTransfer;
} // TWAIN_GetMultiTransfer


int EZTAPI TWAIN_GetHideUI(void)
{
	return bHideUI;
} // TWAIN_GetHideUI

void EZTAPI TWAIN_SetHideUI(int fHide)
{
	bHideUI = (fHide != 0);
} // TWAIN_SetHideUI


void EZTAPI TWAIN_DisableParent(int fYes)
{
    bDisableParent = (fYes != 0);
} // TWAIN_DisableParent

int EZTAPI TWAIN_GetDisableParent(void)
{
    return bDisableParent;
}

///////////////////////////////////////////////////////////////////////
// TWAIN State Changers

int EZTAPI TWAIN_LoadSourceManager(void)
{
	char		szSMDir[128];
	int			cc;
	OFSTRUCT	of;

	if (nState >= TWAIN_SM_LOADED) return TRUE;			// DSM already loaded

	GetWindowsDirectory(szSMDir, sizeof(szSMDir));
	cc = lstrlen(szSMDir);
	if (cc && szSMDir[cc-1] != ':') {
		lstrcat(szSMDir, "\\");
	}
	lstrcat(szSMDir, DSM_FILENAME);			// could crash!
	if (OpenFile(szSMDir, &of, OF_EXIST) != -1) {
		hDSMLib = LoadLibrary(szSMDir);
	} else {
		hDSMLib = NULL;
	}
	if (IsValidHandle(hDSMLib)) {
		pDSM_Entry = (DSMENTRYPROC) GetProcAddress(hDSMLib, DSM_ENTRYPOINT);
		if (pDSM_Entry) {
			iAvailable = 1;
			SetState(TWAIN_SM_LOADED);
		} else {
			FreeLibrary(hDSMLib);
			hDSMLib = NULL;
		}
	} else {
		pDSM_Entry = NULL;
	}
	
	return (nState >= TWAIN_SM_LOADED);
} // TWAIN_LoadSourceManager


static TW_INT32 hwnd32SM;

int EZTAPI TWAIN_OpenSourceManager(HWND hwnd)
{
	hwnd32SM = (TW_INT32)GetValidHwnd(hwnd);

   if (nState < TWAIN_SM_OPEN) {
	   if (TWAIN_LoadSourceManager() &&
		   TWAIN_Mgr(DG_CONTROL, DAT_PARENT, MSG_OPENDSM, &hwnd32SM)) {
		   assert(nState == TWAIN_SM_OPEN);
	   }
   }

	return (nState >= TWAIN_SM_OPEN);
} // TWAIN_OpenSourceManager



int EZTAPI TWAIN_OpenDefaultSource(void)
{
    static TW_IDENTITY twid;
    if (nState < TWAIN_SOURCE_OPEN) {
        if (nState < TWAIN_SM_OPEN && !TWAIN_OpenSourceManager(NULL)) {
            return FALSE;
        }
        
        // open the system default source
        FMEMSET(&twid, 0, sizeof twid);
        if (TWAIN_Mgr(DG_CONTROL, DAT_IDENTITY, MSG_OPENDS, &twid)) {
            assert(nState == TWAIN_SOURCE_OPEN);
        }
    }
    
    return (nState == TWAIN_SOURCE_OPEN);
} // TWAIN_OpenDefaultSource


int EZTAPI TWAIN_GetDefaultSourceName(LPSTR pzName)
{
    TW_IDENTITY twid;
    if (!TWAIN_OpenSourceManager(NULL) ||
        !TWAIN_Mgr(DG_CONTROL, DAT_IDENTITY, MSG_GETDEFAULT, &twid)) {
        return FALSE;
    }
    strcpy(pzName, twid.ProductName);
    return TRUE;
} // TWAIN_GetDefaultSourceName


static unsigned uSLMsg = MSG_GETFIRST;

int EZTAPI TWAIN_GetSourceList(void)
{
    if (!TWAIN_OpenSourceManager(NULL)) {
        TWAIN_ReportLastError("Unable to open TWAIN Source Manager");
        return 0;
    }
    uSLMsg = MSG_GETFIRST;
    return 1;
} // TWAIN_GetSourceList


int EZTAPI TWAIN_GetNextSourceName(LPSTR pzName)
{
    TW_IDENTITY twid;
    memset(&twid, 0, sizeof twid);
    *pzName = '\0';
    if (!TWAIN_Mgr(DG_CONTROL, DAT_IDENTITY, uSLMsg, &twid)) {
        uSLMsg = MSG_GETFIRST;
        return FALSE;
    }
    uSLMsg = MSG_GETNEXT;
    strcpy(pzName, twid.ProductName);
    return TRUE;
} // TWAIN_GetNextSourceName


int EZTAPI TWAIN_OpenSource(LPCSTR pzName)
{
    if (nState >= 4) {
        // some source already open
        if (_stricmp(SourceId.ProductName, pzName) != 0) {
            // some other source open
            TWAIN_CloseSource();
        }
    }

    if (nState < 4) {
        // Prevent TWAIN (or some DS?) from clobbering the current dir.
        // Per David Offen 3/7/00
        TCHAR szDir[_MAX_PATH+1];
        GetCurrentDirectory(_MAX_PATH, szDir);

        if (!pzName || !*pzName) {
            // No specific DS specified - open default
            TWAIN_OpenDefaultSource();
        } else if (!TWAIN_OpenSourceManager(NULL)) {
            TWAIN_ReportLastError("Unable to open TWAIN Source Manager");
        } else {
            // Look for matching DS
            unsigned uMsg = MSG_GETFIRST;
            BOOL bFound = FALSE;
            TW_IDENTITY twid;
            while (memset(&twid, 0, sizeof twid) &&
                   TWAIN_Mgr(DG_CONTROL, DAT_IDENTITY, uMsg, &twid)) {
                if (_stricmp(twid.ProductName, pzName) == 0) {
                    // bingo
                    bFound = TRUE;
                    break;
                }
                uMsg = MSG_GETNEXT;
            } // while
            if (!bFound) {
                // no match
                wsprintf(szMsg, "No such TWAIN device: %s", pzName);
                TWAIN_ReportLastError(szMsg);
            } else if (!TWAIN_Mgr(DG_CONTROL, DAT_IDENTITY, MSG_OPENDS, &twid)) {
                wsprintf(szMsg, "Unable to open: %s", pzName);
                TWAIN_ReportLastError(szMsg);
            }
        }
        SetCurrentDirectory(szDir);
    }
    return nState >= 4;
} // TWAIN_OpenSource


char* EZTAPI TWAIN_SourceName(void)
{
    return SourceId.ProductName;
} // TWAIN_SourceName

void EZTAPI TWAIN_GetSourceName(LPSTR pzName)
{
    strcpy(pzName, SourceId.ProductName);
} // TWAIN_GetSourceName


int EZTAPI TWAIN_GetSourceIdentity(LPVOID ptwid)
{
    if (ptwid) {
        FMEMCPY(ptwid, &SourceId, sizeof SourceId);
    }
    return sizeof SourceId;
} // TWAIN_GetSourceIdentity


int EZTAPI TWAIN_EnableSource(HWND hwnd)
{
   if (nState < TWAIN_SOURCE_OPEN && !TWAIN_OpenDefaultSource()) {
      return FALSE;
   }

	twUI.ShowUI = (TW_BOOL)!bHideUI;
	twUI.hParent = (TW_HANDLE)GetValidHwnd(hwnd);
	TWAIN_DS(DG_CONTROL, DAT_USERINTERFACE, MSG_ENABLEDS, &twUI);
	return (nState == TWAIN_SOURCE_ENABLED);
} // TWAIN_EnableSource


int EZTAPI TWAIN_DisableSource(void)
{
    TWAIN_BreakModalLoop();
    TWAIN_AbortAllPendingXfers();
	if (nState >= TWAIN_SOURCE_ENABLED &&
		TWAIN_DS(DG_CONTROL, DAT_USERINTERFACE, MSG_DISABLEDS, &twUI)) {
		assert(nState == TWAIN_SOURCE_OPEN);
	}
    TWAIN_EmptyMessageQueue();
	return (nState < TWAIN_SOURCE_ENABLED);
} // TWAIN_DisableSource


int EZTAPI TWAIN_CloseSource(void)
{
	rc = TWRC_SUCCESS;

    TWAIN_BreakModalLoop();
	TWAIN_DisableSource();
	if (nState == TWAIN_SOURCE_OPEN &&
		TWAIN_Mgr(DG_CONTROL, DAT_IDENTITY, MSG_CLOSEDS, &SourceId)) {
		assert(nState == TWAIN_SM_OPEN);
	}
	return (nState <= TWAIN_SM_OPEN);
} // TWAIN_CloseSource


int EZTAPI TWAIN_CloseSourceManager(HWND hwnd)
{
    TWAIN_EmptyMessageQueue();
	hwnd32SM = (TW_INT32)GetValidHwnd(hwnd);

	rc = TWRC_SUCCESS;

	if (TWAIN_CloseSource() &&
		TWAIN_Mgr(DG_CONTROL, DAT_PARENT, MSG_CLOSEDSM, &hwnd32SM)) {
		assert(nState == TWAIN_SM_LOADED);
	}
	return (nState <= TWAIN_SM_LOADED);
} // TWAIN_CloseSourceManager


int EZTAPI TWAIN_UnloadSourceManager(void)
{
    TWAIN_CloseSourceManager(NULL);
    if (hwndProxy) {
        DestroyWindow(hwndProxy);
        hwndProxy = NULL;
    }
    if (nState == TWAIN_SM_LOADED) {
        if (hDSMLib) {
            FreeLibrary(hDSMLib);
            hDSMLib = NULL;
        }
        pDSM_Entry = NULL;
        SetState(TWAIN_PRESESSION);
    }
    return (nState == TWAIN_PRESESSION);
} // TWAIN_UnloadSourceManager


///////////////////////////////////////////////////////////////////////
// EZTWAIN's private message pump & support routines
///////////////////////////////////////////////////////////////////////

void EZTAPI TWAIN_ModalEventLoop(void)
{
    MSG msg;
    // Clear global breakout flag
    bBreakModalLoop = FALSE;
    bCloseDsRequested = FALSE;
    
    while ((nState >= TWAIN_SOURCE_ENABLED)
        && !bBreakModalLoop
        && GetMessage((LPMSG)&msg, NULL, 0, 0)) {
        
        if (!TWAIN_MessageHook ((LPMSG)&msg)) {
            TranslateMessage ((LPMSG)&msg);
            DispatchMessage ((LPMSG)&msg);
        }
    } // while
    bBreakModalLoop = FALSE;
} // TWAIN_ModalEventLoop


void EZTAPI TWAIN_BreakModalLoop(void)
{
   bBreakModalLoop = TRUE;
} // TWAIN_BreakModalLoop


void EZTAPI TWAIN_EmptyMessageQueue(void)
{
    MSG msg;
#ifdef _DEBUG
    OutputDebugString("EmptyMsgQ<");
#endif
    while (PeekMessage((LPMSG)&msg, NULL, 0, 0, PM_REMOVE)) {
        if (msg.message == WM_QUIT) {
            PostQuitMessage(msg.wParam);
            break;
        }
        if (!TWAIN_MessageHook((LPMSG)&msg)) {
        	TranslateMessage((LPMSG)&msg);
        	DispatchMessage((LPMSG)&msg);
#ifdef _DEBUG
            OutputDebugString("-");
#endif
        } else {
#ifdef _DEBUG
            OutputDebugString("T");
#endif
        }
    }
#ifdef _DEBUG
    OutputDebugString(">\n");
#endif
} // TWAIN_EmptyMessageQueue


///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
// Specialists for each of the three transfer modes

///////////////////////////////////////////////////////////////////////
// Native Transfer Mode transfer handler

static void NativeXferHandler(void)
{
    TW_UINT32		hNative;
    
    assert(nState == TWAIN_TRANSFER_READY);
    if (TWAIN_DS(DG_IMAGE, DAT_IMAGENATIVEXFER, MSG_GET, &hNative)) {
        hDib = (HANDLE)hNative;
    } else {
        hDib = NULL;
        if (TWAIN_GetResultCode() == TWRC_CANCEL) {
            RecordError(EZTEC_USER_CANCEL);
        } else {
            RecordError(EZTEC_XFER_FAILURE);
        }
    }
} // NativeXferHandler


///////////////////////////////////////////////////////////////////////
// File Transfer Mode transfer handler
static void FileXferHandler(void)
{
    TW_SETUPFILEXFER twsfx;
    twsfx.Format = (TW_UINT16)nFileXferFormat;
    strncpy(twsfx.FileName, szFileXferPath, 255);
    if (TWAIN_DS(DG_CONTROL, DAT_SETUPFILEXFER, MSG_SET, &twsfx) &&
        TWAIN_DS(DG_IMAGE, DAT_IMAGEFILEXFER, MSG_GET, NULL)) {
        bXferDone = TRUE;
    } else if (TWAIN_GetResultCode() == TWRC_CANCEL) {
        RecordError(EZTEC_USER_CANCEL);
    } else {
        RecordError(EZTEC_XFER_FAILURE);
    }
} // FileXferHandler


///////////////////////////////////////////////////////////////////////
// Memory Transfer Mode transfer handler (and support routines)

static BOOL MemXferReady(TW_IMAGEINFO info)
{
    int bpp = info.BitsPerPixel;
    assert(hDib == NULL);
    if (info.ImageLength < 0) {
        // can't handle unknown length
        TWAIN_ErrorBox("EZTW: Unable to transfer unknown image length");
        return FALSE;
    }
    if (info.ImageWidth < 0) {
        // can't handle unknown width
        TWAIN_ErrorBox("EZTW: Unable to transfer unknown image width.");
        return FALSE;
    }
    if (info.Compression != TWCP_NONE) {
        TWAIN_ErrorBox("EZTW: Unable to transfer compressed memory data.");
        return FALSE;
    }
    if (info.PixelType > TWPT_PALETTE) {
        TWAIN_ErrorBox("EZTW: Unable to transfer exotic pixel type: CMY,YUV,XYZ,...");
        return FALSE;
    }
    hDib = DIB_Allocate(bpp, info.ImageWidth, info.ImageLength);
    if (!hDib) {
        TWAIN_ErrorBox("EZTW: Unable to allocate memory for image.");
        return FALSE;
    }
    DIB_SetResolution(hDib, Fix32ToFloat(info.XResolution), Fix32ToFloat(info.YResolution));
    return TRUE;
} // MemXferReady


static int bReported1;

static void PutOnClipboard(const char* pz)
{
    if (OpenClipboard(NULL)) {
        char szText[1024], *q = szText;
        HANDLE h;
        while (*pz) {
            char ch = *pz++;
            if (ch == '\n') {
                *q++ = '\r'; *q++ = '\n';
            } else {
                *q++ = ch;
            }
        }
        *q = '\0';
        h = GlobalAlloc(GMEM_MOVEABLE | GMEM_DDESHARE, strlen(szText)+1);
        if (h) {
            q = (char *)GlobalLock(h);
            if (q) {
                strcpy(q, szText);
                GlobalUnlock(h);
                SetClipboardData(CF_TEXT, h);
            }
        }
        CloseClipboard();
    }
} // PutOnClipboard


static void DisplayTransferInfo(const char* pzNote, TW_IMAGEMEMXFER mxb,
                                TW_IMAGEINFO info,
                                int nChannel)
{
    if (!bReported1) {
        char szText[1024];
        sprintf(szText,
            "EZTWAIN %d.%02d: Memory Transfer Incompatibility\n"
            "(%s)\n\n"
            "This text has been copied to the clipboard.\n"
            "Please paste it into an e-mail to: support@dosadi.com\n\n"
            "Product \"%s\"\n"
            "Version %u-%u  Vendor: \"%s\"\n"
            "Channel: %d   MXB.Memory.Flags: 0x%x\n"
            "MXB.Comp: %u, BytesPerRow: %ld, BytesWritten: %ld\n"
            "MXB.Columns: %ld, Rows: %ld, XOffset: %ld, YOffset: %ld\n"
            "INFO.ImageWidth: %ld, ImageLength: %ld\n"
            "INFO.SPP: %d, BPS: [%d, %d, %d, %d], BPP: %d\n"
            "INFO.Planar: %d, PixelType: %d, Compression: %u\n",
            VERSION / 100, VERSION % 100,
            pzNote,
            (char*)SourceId.ProductName,
            (unsigned)SourceId.Version.MajorNum,
            (unsigned)SourceId.Version.MinorNum,
            (char*)SourceId.Manufacturer,
            (int)nChannel,
            (unsigned)mxb.Memory.Flags,
            (unsigned)mxb.Compression,
            (long)mxb.BytesPerRow,
            (long)mxb.BytesWritten,
            (long)mxb.Columns,
            (long)mxb.Rows,
            (long)mxb.XOffset,
            (long)mxb.YOffset,
            (long)info.ImageWidth,
            (long)info.ImageLength,
            (int)info.SamplesPerPixel,
            (int)info.BitsPerSample[0],
            (int)info.BitsPerSample[1],
            (int)info.BitsPerSample[2],
            (int)info.BitsPerSample[3],
            (int)info.BitsPerPixel,
            (int)info.Planar,
            (int)info.PixelType,
            (unsigned)info.Compression
            );

        if (pfsLog) {
           fprintf(pfsLog,
              "DisplayTransferInfo - message follows -\n%s------------------------",
              szText);
        }

        PutOnClipboard(szText);
        TWAIN_ErrorBox(szText);
        bReported1 = TRUE;
    }
} // DisplayTransferInfo


static void LogInfo(TW_IMAGEINFO info)
{
    fprintf(pfsLog,
        "INFO: ImageWidth:  %4ld, ImageLength: %4ld\n"
        "      PixelType: %d  Planar: %d, Compression: %u\n"
        "      BitsPerPixel: %2d, SamplesPerPixel: %d, BitsPerSample: [%d, %d, %d, %d]\n"
        "      n",
        (long)info.ImageWidth,
        (long)info.ImageLength,
        (int)info.PixelType,
        (int)info.Planar,
        (unsigned)info.Compression,
        (int)info.BitsPerPixel,
        (int)info.SamplesPerPixel,
        (int)info.BitsPerSample[0],
        (int)info.BitsPerSample[1],
        (int)info.BitsPerSample[2],
        (int)info.BitsPerSample[3]
        );
} // LogInfo


static void LogMXB(TW_IMAGEMEMXFER mx)
{
    fprintf(pfsLog,
        "MXB: Memory.Flags: 0x%x Comp: %u, BytesPerRow: %ld, BytesWritten: %ld\n"
        "     Columns: %ld, Rows: %ld, XOffset: %ld, YOffset: %ld\n",
        (unsigned)mx.Memory.Flags,
        (unsigned)mx.Compression,
        (long)mx.BytesPerRow,
        (long)mx.BytesWritten,
        (long)mx.Columns,
        (long)mx.Rows,
        (long)mx.XOffset,
        (long)mx.YOffset
        );
} // LogMXB


BOOL WINAPI MemoryXferHandler(LPVOID pdata)
{
    TW_IMAGEINFO    info;
    TW_SETUPMEMXFER twsx;
    TW_IMAGEMEMXFER mx;
    BOOL            bDone = FALSE;
    int             i;
    int             nChannel = 0;          // for multi-plane
    int             nPlanes;
    int             nPixels;

    pdata = pdata;              // suppress unreferenced parameter warning
    
    if (pfsLog) fprintf(pfsLog, "start of MemoryXferHandler\n");
    
    assert(hDib == NULL);
    assert(nState == TWAIN_TRANSFER_READY);
    
    __try {
       if (!TWAIN_DS(DG_CONTROL, DAT_SETUPMEMXFER, MSG_GET, &twsx)) {
          if (pfsLog) fprintf(pfsLog, "  DAT_SETUPMEMXFER, MSG_GET failed\n"); 
          return FALSE;
       }
       // Get actual image info
       if (!TWAIN_DS(DG_IMAGE, DAT_IMAGEINFO, MSG_GET, &info)) {
           if (pfsLog) fprintf(pfsLog,
               "  DAT_IMAGEINFO, MSG_GET failed!\n"
               "  aborting transfer\n");
           return FALSE;
       }
       if (pfsLog) LogInfo(info);
       // Compute the number of pixels in one plane of this image:
       nPixels = info.ImageWidth * info.ImageLength;
       // Figure out how many planes are in this image:
       nPlanes = info.Planar ? info.SamplesPerPixel : 1;
       if (nPlanes != 1 && pfsLog) fprintf(pfsLog, "Channel = 0\n");
       
       // Time to allocate the DIB
       if (!MemXferReady(info)) {
           // Unable or unwilling to receive this image, abort the transfer.
           if (pfsLog) fprintf(pfsLog,
               "  MemXferReady(info) returned false!\n"
               "  aborting transfer\n");
           return FALSE;
       }
       assert(hDib != NULL);
       // Allocate memory buffer
       mx.Memory.TheMem = (LPVOID)malloc(twsx.Preferred);
       if (!mx.Memory.TheMem) {
           if (pfsLog) fprintf(pfsLog, "  malloc(twsx.Preferred) returned NULL\n"
                                       "  aborting transfer.\n"); 
          TWAIN_FreeNative(hDib); hDib = NULL;
          TWAIN_ErrorBox("EZTW: Transfer buffer allocation failed.");
          return FALSE;
       }
       mx.Memory.Length = twsx.Preferred;
       mx.Memory.Flags = TWMF_APPOWNS | TWMF_POINTER;
       
       while (!bDone && TWAIN_DS(DG_IMAGE, DAT_IMAGEMEMXFER, MSG_GET, &mx)) {
          if (pfsLog) LogMXB(mx);

          if (mx.BytesWritten > twsx.Preferred) {
              // yeah, right
             DisplayTransferInfo("mx.BytesWritten > twsx.Preferred", mx, info, nChannel);
             break;
          }
          if (mx.XOffset == 0 && (TW_INT32)mx.Columns > info.ImageWidth) {
              // HP PictureScan 3.0 hack
              if (mx.BytesWritten > 0 && mx.BytesPerRow * mx.Rows == mx.BytesWritten) {
                  // the 'stride' info looks OK, just clip the data to the info
                  mx.Columns = info.ImageWidth;
              }
          }
          // Try to write the data into the DIB
          if (!CopyMemoryBufferToDib(hDib, mx, info, nChannel)) {
             // DIB writer says to abort
             DisplayTransferInfo("CopyMemoryBufferToDib failed", mx, info, nChannel);
             break;
          }
          nPixels -= (mx.Columns * mx.Rows);
          if (TWAIN_GetResultCode() == TWRC_XFERDONE) {
             // DS says that was the last buffer:
             if (pfsLog) {
                fprintf(pfsLog, "  RC = TWRC_XFERDONE, ending transfer\n");
                fprintf(pfsLog, "    nPixels = %d\n", nPixels);
             }
             bDone = TRUE;
          } else if (nPixels <= 0) {
             // Finished with one plane of image
             nChannel++;
             if (nChannel >= nPlanes) {
                // All planes done
                if (pfsLog) {
                   fprintf(pfsLog, "%d Planes filled, ending transfer\n", nChannel);
                }
                bDone = TRUE;
             } else {
                assert(info.Planar);
                nPixels = info.ImageWidth * info.ImageLength;
                if (pfsLog) {
                   fprintf(pfsLog, "Starting Channel %d\n", nChannel);
                }
             }
          }
       } // while
       
       if (!bDone && TWAIN_GetResultCode() == TWRC_CANCEL) {
          if (pfsLog) {
             fprintf(pfsLog, "  last RC = TWRC_CANCEL, RecordError.\n");
          }
          RecordError(EZTEC_USER_CANCEL);
       }
       if (bDone && hDib) {
           // Do various post-transfer stuff to DIB
           if (DIB_Depth(hDib) == 24 && !info.Planar) {
               // 24-bit memory transfer - have to swap RGB -> BGR in Dibs
               if (pfsLog) {
                   fprintf(pfsLog, "24-bit non-Planar: BmiBuggerAll.\n");
               }
               DIB_SwapRedBlue(hDib);
           }
           
           if (info.PixelType == TWPT_PALETTE) {
               // Get the palette for the image
               TW_PALETTE8 pal8;
               if (!TWAIN_DS(DG_IMAGE, DAT_PALETTE8, MSG_GET, &pal8)) {
                   if (pfsLog) {
                       fprintf(pfsLog, "  DAT_PALETTE8, MSG_GET failed\n"); 
                   }
                   TWAIN_FreeNative(hDib); hDib = NULL;
                   TWAIN_ErrorBox("EZTW: Can't get image palette.");
                   return FALSE;
               }
               DIB_SetColorCount(hDib, pal8.NumColors);
               for (i = 0; i < pal8.NumColors; i++) {
                   int e = pal8.Colors[i].Index;
                   int r = pal8.Colors[i].Channel1;
                   int g = pal8.Colors[i].Channel2;
                   int b = pal8.Colors[i].Channel3;
                   DIB_SetColorTableRGB(hDib, e, r, g, b);
               } // for

           } // if palette image

       } // if (hDib)

    } // SEH try block
    __finally {             // termination handler
       // This code is invoked even if the try block crashes (supposedly)
       if (pfsLog) {
          fprintf(pfsLog, "__finally.  bDone = %s\n", bDone ? "true" : "false"); 
       }
       if (!bDone && hDib) {
          // Something went wrong - free the DIB
          RecordError(EZTEC_XFER_FAILURE);
          TWAIN_FreeNative(hDib); hDib = NULL;
       }
       // Free the transfer buffer
       free(mx.Memory.TheMem);
    }
    if (pfsLog) {
       fprintf(pfsLog, "MemoryXferHandler returning TRUE\n");
    }
    return bDone;
} // MemoryXferHandler


static void DoOneTransfer(void)
{
    switch (nXferMech) {
    case XFERMECH_NATIVE:
        NativeXferHandler();
        break;
    case XFERMECH_FILE:
        FileXferHandler();
        break;
    case XFERMECH_MEMORY:
        assert(nState == TWAIN_TRANSFER_READY);
        bCBSuccess = (*cbXFunc)(cbData);
        cbXFunc = MemoryXferHandler; // only valid this once
        break;
    } // switch
    
    // If inside ModalEventLoop, break out
    TWAIN_BreakModalLoop();
    
    // Acknowledge transfer
    TWAIN_EndXfer();
    assert(nState == TWAIN_TRANSFER_READY ||
        nState == TWAIN_SOURCE_ENABLED);
} // DoOneTransfer


int EZTAPI TWAIN_MessageHook(LPMSG lpmsg)
// returns TRUE if msg processed by TWAIN (source)
{
    int   bProcessed = FALSE;
    
    if (nState >= TWAIN_SOURCE_ENABLED) {
        // source enabled
        TW_EVENT	twEvent;
        twEvent.pEvent = (TW_MEMREF)lpmsg;
        twEvent.TWMessage = MSG_NULL;
        // see if source wants to process (eat) the message
        bProcessed = TWAIN_DS(DG_CONTROL, DAT_EVENT, MSG_PROCESSEVENT, &twEvent);
        switch (twEvent.TWMessage) {
        case MSG_XFERREADY:
            assert(nState == TWAIN_TRANSFER_READY);
            DoOneTransfer();
            break;
        case MSG_CLOSEDSREQ:
            bCloseDsRequested = TRUE;
            TWAIN_DisableSource();
            break;
        case MSG_NULL:
            // no message returned from DS
            break;
        } // switch
    }
    return bProcessed;
} // TWAIN_MessageHook


int EZTAPI TWAIN_EndXfer(void)
{
    TWAIN_BreakModalLoop();
    if (nState == TWAIN_TRANSFERRING) {
        TWAIN_DS(DG_CONTROL, DAT_PENDINGXFERS, MSG_ENDXFER, &pendingXfers);
    }
    return (nState < TWAIN_TRANSFERRING);
} // TWAIN_EndXfer


int EZTAPI TWAIN_AbortAllPendingXfers(void)
{
    TWAIN_BreakModalLoop();
    TWAIN_EndXfer();
    if (nState == TWAIN_TRANSFER_READY) {
        TWAIN_DS(DG_CONTROL, DAT_PENDINGXFERS, MSG_RESET, &pendingXfers);
    }
    TWAIN_EmptyMessageQueue();
    return (nState < TWAIN_TRANSFER_READY);
} // TWAIN_AbortAllPendingXfers


///////////////////////////////////////////////////////////////////////
// File I/O

int EZTAPI TWAIN_IsJpegAvailable(void)
{
#ifdef IJL_JPEG
    return LoadJpegSupport();
#else
    return FALSE;
#endif
} // TWAIN_IsJpegAvailable


int EZTAPI TWAIN_SetSaveFormat(int nFF)
{
    if (nFF == TWFF_BMP || nFF == TWFF_JFIF) {
        if (nFF == TWFF_JFIF && !TWAIN_IsJpegAvailable()) {
            TWAIN_ErrorBox("Cannot save JPEG files: IJL11.DLL or IJL10.DLL not found.");
            return FALSE;
        }
        nSaveFormat = nFF;
        return TRUE;
    }
    return FALSE;
} // TWAIN_SetSaveFormat


int EZTAPI TWAIN_GetSaveFormat(void)
{
    return nSaveFormat;
} // TWAIN_GetSaveFormat


int EZTAPI TWAIN_WriteDibToFile(LPBITMAPINFOHEADER lpDIB, HFILE fh)
{
   return BmiWriteToBmpFile(lpDIB, fh);
} // TWAIN_WriteDibToFile


int EZTAPI TWAIN_WriteNativeToFile(HANDLE hdib, HFILE fh)
{
   return DIB_WriteToBmpFile(hdib, fh);
} //  TWAIN_WriteNativeToFile


BOOL PromptForFilename(char* pszFile, int* pnFF)
// TRUE if OK, FALSE if cancelled.
{
    // prompt for filename
    OPENFILENAME ofn;
    int nExt;
    
    FMEMSET(&ofn, 0, sizeof ofn);
    pszFile[0] = '\0';
    ofn.lStructSize = sizeof(OPENFILENAME);
    ofn.hwndOwner = NULL;
    if (TWAIN_IsJpegAvailable()) {
        ofn.lpstrFilter = "Windows Bitmap (*.bmp)\0*.bmp\0JPEG/JFIF (*.jpg)\0*.jpg\0\0";
    } else {
        ofn.lpstrFilter = "Windows Bitmap (*.bmp)\0*.bmp\0\0";
    }
    if (*pnFF == TWFF_JFIF && TWAIN_IsJpegAvailable()) {
        ofn.lpstrDefExt = "jpg";
        ofn.nFilterIndex = 2;
    } else {
        ofn.lpstrDefExt = "bmp";
        ofn.nFilterIndex = 1;
    }
    ofn.lpstrFile= pszFile;
    ofn.nMaxFile = _MAX_PATH - 5;
    ofn.Flags = OFN_OVERWRITEPROMPT | OFN_HIDEREADONLY |
        OFN_NOREADONLYRETURN;
    
    if (!GetSaveFileName(&ofn)) {
        return FALSE;					// user cancelled dialog
    }
    // Use file format chosen by user in dialog
    *pnFF = (ofn.nFilterIndex == 1) ? TWFF_BMP : TWFF_JFIF;
    // supply default extension - GetSaveFileName doesn't seem to do it!
    nExt = ofn.nFileExtension;
    if (nExt==0 || !pszFile[nExt]) {
        // no extension
        lstrcat(pszFile, (*pnFF == TWFF_BMP) ? ".bmp" : ".jpg");
    }
    return TRUE;
} // PromptForFilename



int EZTAPI TWAIN_WriteNativeToFilename(HANDLE hdib, LPCSTR pszFile)
{
	int result;
	char szFile[_MAX_PATH];
    char* pszExt;
    int nFF = TWAIN_GetSaveFormat();

    if (pszFile) {
        strcpy(szFile, pszFile);
    } else {
        szFile[0] = '\0';
    }

	if (!szFile[0]) {
        // No filename provided, prompt user.
        // NB User can change save format!
        if (!PromptForFilename(szFile, &nFF)) {
            return -1;
        }
    }

    // Look for the extension
    pszExt = strrchr(szFile, '.');
    if (!pszExt || strchr(pszExt, '\\')) {
        // If not found (none supplied) assume .BMP
        pszExt = szFile + strlen(szFile);
        lstrcat(szFile, (nFF == TWFF_BMP) ? ".bmp" : ".jpg");
    }
    if (nFF == TWFF_JFIF) {
        result = DIB_WriteToJpeg(hdib, szFile);
    } else {
        result = DIB_WriteToBmp(hdib, szFile);
    }
	return result;
} // TWAIN_WriteNativeToFilename


HANDLE EZTAPI TWAIN_LoadNativeFromFilename(LPCSTR pszFile)
{
    return DIB_LoadFromFilename(pszFile);
} // TWAIN_LoadNativeFromFilename


#ifdef IJL_JPEG

static int nJpegQuality = 75;       // "good" quality

void EZTAPI TWAIN_SetJpegQuality(int nQ)
{
    nJpegQuality = nQ;
} // TWAIN_SetJpegQuality

int EZTAPI TWAIN_GetJpegQuality(void)
{
    return nJpegQuality;
} // TWAIN_GetJpegQuality


static int LoadJpegSupport(void)
{
    if (hijl == NULL) {
        hijl = LoadLibrary("ijl11.dll");
        if (hijl == NULL) {
            hijl = LoadLibrary("ijl10.dll");
        }
        if (hijl) {
            pijlInit = (PIJLINIT)GetProcAddress(hijl, "ijlInit");
            pijlFree = (PIJLINIT)GetProcAddress(hijl, "ijlFree");
            pijlWrite = (PIJLIO)GetProcAddress(hijl, "ijlWrite");
            pijlRead = (PIJLIO)GetProcAddress(hijl, "ijlRead");
            if (!pijlInit || !pijlFree || !pijlWrite || !pijlRead) {
                FreeLibrary(hijl); hijl = NULL;
            }
        }
    }
    return (hijl != NULL);
}


static HANDLE LoadJpegFromBuffer(LPBYTE pbuf, long nbytes)
// Try to parse a JPEG/JFIF image from the buffer at pbuf
// containing nbytes bytes of data.
{
    HANDLE hdib = NULL;
    LPBITMAPINFOHEADER pbi;
    JPEG_CORE_PROPERTIES jcprops;
    IJLERR err;
    err = pijlInit(&jcprops);
    if (IJL_OK == err) {
        jcprops.JPGBytes = pbuf;
        jcprops.JPGSizeBytes = nbytes;
        err = pijlRead(&jcprops, IJL_JBUFF_READPARAMS);
        if (IJL_OK == err && jcprops.JPGChannels != 3 && jcprops.JPGChannels != 1) {
            err = IJL_UNSUPPORTED_BYTES_PER_PIXEL;
        }
        if (IJL_OK == err) {
            jcprops.DIBWidth = jcprops.JPGWidth;
            jcprops.DIBHeight = jcprops.JPGHeight;
            jcprops.DIBChannels = jcprops.JPGChannels;
            jcprops.DIBColor = (jcprops.DIBChannels == 3) ? IJL_BGR : IJL_G;
            jcprops.JPGColor = (jcprops.DIBChannels == 3) ? IJL_YCBCR : IJL_G;

            hdib = DIB_Allocate(jcprops.DIBChannels*8, jcprops.DIBWidth, jcprops.DIBHeight);
            pbi = (LPBITMAPINFOHEADER)GlobalLock(hdib);
            if (!hdib || !pbi) {
                err = IJL_MEMORY_ERROR;
            } else {
                jcprops.DIBBytes = BmiFindBits(pbi);
                jcprops.DIBPadBytes = BmiRowBytes(pbi) - (jcprops.DIBWidth * jcprops.DIBChannels);
                jcprops.DIBHeight *= -1;
                err = pijlRead(&jcprops, IJL_JBUFF_READWHOLEIMAGE);
            }
            GlobalUnlock(hdib);
        }
    }
    if (err != IJL_OK) {
        wsprintf(szMsg, "JPEG Library Error %d\n", err);
        TWAIN_ErrorBox(szMsg);
    }
    pijlFree(&jcprops);
    return hdib;
}


static HANDLE LoadJpeg(HFILE fh)
{
    HANDLE hdib = NULL;
    long fileSize = GetFileSize((HANDLE)fh, NULL);//_filelength(fh);
    HANDLE hbuf = GlobalAlloc(GMEM_MOVEABLE, fileSize);
    if (hbuf) {
        LPBYTE pbuf = (LPBYTE)GlobalLock(hbuf);
        if (pbuf) {
            if (fileSize == HUGEREAD(fh, pbuf, fileSize)) {
                hdib = LoadJpegFromBuffer(pbuf, fileSize);
            }
            GlobalUnlock(hbuf);
        }
        GlobalFree(hbuf);
    }
    return hdib;
} // LoadJpeg


int WriteToJpeg(LPBITMAPINFOHEADER pbi, LPCSTR pszFile)
{
    int nResult = -4;
    JPEG_CORE_PROPERTIES jcprops;
    IJLERR err;
    if (!LoadJpegSupport()) {
        TWAIN_ErrorBox("Cannot write JPEG file: IJL11.DLL or IJL10.DLL not found.");
        return -3;
    }

    err = pijlInit(&jcprops);
    if (IJL_OK == err) {
       BOOL bColor = pbi->biBitCount > 8;      // cheap...
       jcprops.DIBBytes = BmiFindBits(pbi);
       jcprops.DIBWidth = pbi->biWidth;
       jcprops.DIBHeight = BmiHeight(pbi);
       jcprops.DIBChannels = pbi->biBitCount / 8;
       jcprops.DIBPadBytes = BmiRowBytes(pbi) - (jcprops.DIBWidth * jcprops.DIBChannels);
       jcprops.DIBColor = bColor ? IJL_BGR : IJL_G;
       jcprops.JPGFile = (LPTSTR)pszFile;
       jcprops.JPGWidth = jcprops.DIBWidth;
       jcprops.JPGHeight = jcprops.DIBHeight;
       jcprops.JPGChannels = jcprops.DIBChannels;
       jcprops.JPGColor = bColor ? IJL_YCBCR : IJL_G;
       jcprops.JPGSubsampling = bColor ? IJL_411 : IJL_NONE;
       jcprops.jquality = nJpegQuality;
       jcprops.DIBHeight *= -1;
       DeleteFile(pszFile);
       err = pijlWrite(&jcprops, IJL_JFILE_WRITEWHOLEIMAGE);
       if (err == IJL_OK) {
          nResult = 0;
       } else if (err == IJL_INVALID_FILENAME) {
          nResult = -2;
       } else if (err == IJL_INVALID_JPEG_PROPERTIES) {
          nResult = -3;   // invalid image format
       } else {
          nResult = -4;
          wsprintf(szMsg, "JPEG Library Write Error %d\n", err);
          TWAIN_ErrorBox(szMsg);
       }
       pijlFree(&jcprops);
    }
    return nResult;
} // WriteToJpeg


#else

int WriteToJpeg(LPBITMAPINFOHEADER pbi, LPCSTR pszFile)
{
    TWAIN_ErrorBox("This version of EZTWAIN does not support JPEG/JFIF files.");
    return -3;
}


#endif // IJL_JPEG


HANDLE EZTAPI TWAIN_LoadNativeFromFile(HFILE fh)
// Like LoadNativeFromFilename, but takes an already open file handle.
{
    int nFF;
	HANDLE hdib;
	LPBYTE pbi;
    BYTE tag[10];
	BITMAPFILEHEADER bmh;
	INT32 dibSize;
	// Read file header and identify
    memset(tag, 0, sizeof tag);
    _lread(fh, tag, sizeof tag);
	if (tag[0] == 0x42 && tag[1] == 0x4D) {
        // Windows BMP file
        nFF = TWFF_BMP;
#ifdef IJL_JPEG
    } else if (tag[0] == 0xFF && tag[1] == 0xD8 && tag[2] == 0xFF && tag[3] == 0xE0 &&
               tag[6] == 'J' && tag[7] == 'F' && tag[8] == 'I' && tag[9] == 'F') {
        // JPEG in JFIF form
        nFF = TWFF_JFIF;
#endif
    } else {
        // Not a recognized file format
		return NULL;
	}
    // Rewind to start of file for loading
    //_lseek(fh, 0L, SEEK_SET);
    SetFilePointer((HANDLE)fh, 0, NULL, FILE_BEGIN);
    //
    switch (nFF) {
    case TWFF_BMP:
        // Read BMP file header
        if (_lread(fh, &bmh, sizeof bmh) != sizeof bmh) {
            return NULL;
        }
        // Allocate global block for DIB
        dibSize = bmh.bfSize - sizeof bmh;
        hdib = GlobalAlloc(GMEM_MOVEABLE, dibSize);
        pbi = (LPBYTE)GlobalLock(hdib);
        if (!hdib || !pbi) {
            return NULL;
        }
        
        // Read DIB from file
        if (HUGEREAD(fh, pbi, dibSize) != dibSize) {
            GlobalUnlock(hdib);
            GlobalFree(hdib);
            return NULL;
        }
        GlobalUnlock(hdib);
        break;

#ifdef IJL_JPEG
    case TWFF_JFIF:
        if (!LoadJpegSupport()) {
            TWAIN_ErrorBox("Cannot read JPEG file: IJL11.DLL or IJL10.DLL not found.");
            return NULL;
        }
        hdib = LoadJpeg(fh);
        break;
#endif

    } // switch
	return hdib;
} // TWAIN_LoadNativeFromFile

///////////////////////////////////////////////////////////////////////
// TWAIN State 4 Negotiation Functions

int EZTAPI TWAIN_NegotiateXferCount(int nXfers)
{
	return TWAIN_SetCapOneValue(CAP_XFERCOUNT, TWTY_INT16, nXfers);
} // TWAIN_NegotiateXferCount


int EZTAPI TWAIN_NegotiatePixelTypes(unsigned wPixTypes)
{
    TW_CAPABILITY 		cap;
    void far *			pv;
    int					fSuccess = FALSE;
    
    if (nState != TWAIN_SOURCE_OPEN) {
        return RecordError(EZTEC_NOT_STATE_4);
    }
    
    if (TWAIN_ANYTYPE == wPixTypes) {
        // Any pixel type is OK
        return TRUE;			// that was easy!
    }

    // Fill in capability structure
    cap.Cap = ICAP_PIXELTYPE;			// capability id
    cap.ConType = TWON_DONTCARE16;

    if (!TWAIN_DS(DG_CONTROL, DAT_CAPABILITY, MSG_GET, (TW_MEMREF)&cap)) {
        return RecordError(EZTEC_CAP_GET);
    }
    if (!cap.hContainer) {
        return RecordError(EZTEC_NULL_HCON);
    }
    pv = GlobalLock(cap.hContainer);
    if (!pv) {
        // this source is invalid, further negotiation is unlikely to succeed
        return RecordError(EZTEC_BAD_HCON);
    }
    
    switch (cap.ConType) {
        
    case TWON_ENUMERATION:
        {
            TW_ENUMERATION far *pcon = (TW_ENUMERATION far *)pv;
            if (pcon->NumItems < 1) {
                RecordError(EZTEC_CAP_GET_EMPTY);
            } else if (pcon->ItemType != TWTY_UINT16 && pcon->ItemType != TWTY_INT16) {
                // We can only handle the two types
                RecordError(EZTEC_BAD_ITEMTYPE);
            } else {
                TW_UINT16 FAR * plist = (TW_UINT16 FAR *)pcon->ItemList;
                TW_UINT32 nItems = pcon->NumItems;
                TW_UINT32 nCurrent = pcon->CurrentIndex;
                TW_UINT32 nDefault = pcon->DefaultIndex;
                TW_UINT16 uCurrentVal = plist[nCurrent < nItems ? nCurrent : 0];
                TW_UINT16 uDefaultVal = plist[nDefault < nItems ? nDefault : 0];
                nItems = Intersect(wPixTypes, nItems, plist);
                pcon->NumItems = nItems;
                pcon->CurrentIndex = FindIndex16(nItems, plist, uCurrentVal, 0);
                pcon->DefaultIndex = FindIndex16(nItems, plist, uDefaultVal, 0);
                fSuccess = (pcon->NumItems != 0);
                if (!fSuccess) RecordError(EZTEC_CAP_SET_EMPTY);
            }
            break;
        }
        
    case TWON_ARRAY:
        {
            // this is technically illegal - TWAIN 1.5, p9-30, Containers for MSG_GET: TW_ENUMERATION, TW_ONEVALUE
            TW_ARRAY far *pcon = (TW_ARRAY far *)pv;
            if (pcon->NumItems < 1) {
                RecordError(EZTEC_CAP_GET_EMPTY);
            } else if (pcon->ItemType != TWTY_UINT16 && pcon->ItemType != TWTY_INT16) {
                RecordError(EZTEC_BAD_ITEMTYPE);
            } else {
                pcon->NumItems = Intersect(wPixTypes, (unsigned)pcon->NumItems, (TW_UINT16 far *)pcon->ItemList);
                fSuccess = (pcon->NumItems != 0);
                if (!fSuccess) RecordError(EZTEC_CAP_SET_EMPTY);
            }
            break;
        }
        
    case TWON_ONEVALUE:
        {
            TW_ONEVALUE far *pcon = (TW_ONEVALUE far *)pv;
            fSuccess = ((1 << pcon->Item) & wPixTypes);
            if (!fSuccess) RecordError(EZTEC_CAP_SET_EMPTY);
            break;
        }
        
    default:
        // something we don't understand, abandon negotiations
        RecordError(EZTEC_BAD_CONTYPE);
        break;
    } // switch
    
    GlobalUnlock(cap.hContainer);
    
    if (fSuccess) {
        // For enums (and arrays) send intersection back, to restrict it.
        // For one vals, don't bother - could only cause confusion.
        if (cap.ConType != TWON_ONEVALUE) {
            fSuccess = TWAIN_DS(DG_CONTROL, DAT_CAPABILITY, MSG_SET, (TW_MEMREF)&cap);
            if (!fSuccess) RecordError(EZTEC_CAP_SET);
        }
    }
    GlobalFree(cap.hContainer);
    
    if (fSuccess) {
#ifdef _DEBUG
        OutputDebugString("EZTW:Pixel Types negotiated.\n");
#endif
    }
    return fSuccess;
} // TWAIN_NegotiatePixelTypes


int EZTAPI TWAIN_GetCurrentUnits(void)
{
    return TWAIN_GetCapUint16(ICAP_UNITS, TWUN_INCHES);
} // TWAIN_GetCurrentUnits


int EZTAPI TWAIN_SetCurrentUnits(int nUnits)
// Negotiate the current pixel type for acquisition.
// Negotiation is only allowed in State 4 (TWAIN_SOURCE_OPEN)
// The source may select this pixel type, but don't assume it will.
{
	return TWAIN_SetCapOneValue(ICAP_UNITS, TWTY_UINT16, (TW_UINT16)nUnits);
} // TWAIN_SetCurrentUnits


int EZTAPI TWAIN_GetBitDepth(void)
// Ask the source for the current bitdepth.
// This value depends on the current PixelType.
// Bit depth is per color channel e.g. 24-bit RGB has bit depth 8.
// If anything goes wrong, this function returns 0.
{
    return TWAIN_GetCapUint16(ICAP_BITDEPTH, 0);
} // TWAIN_GetBitDepth

int EZTAPI TWAIN_SetBitDepth(int nBits)
// (Try to) set the current bitdepth (for the current pixel type).
{
	return TWAIN_SetCapOneValue(ICAP_BITDEPTH, TWTY_UINT16, (TW_UINT16)nBits);
} // TWAIN_SetBitDepth


int EZTAPI TWAIN_GetPixelType(void)
// Ask the source for the current pixel type.
// If anything goes wrong (it shouldn't!), this function returns 0 (TWPT_BW).
{
    return TWAIN_GetCapUint16(ICAP_PIXELTYPE, 0);
} // TWAIN_GetPixelType

int EZTAPI TWAIN_SetCurrentPixelType(int nPixType)
// Negotiate the current pixel type for acquisition.
// Negotiation is only allowed in State 4 (TWAIN_SOURCE_OPEN)
// The source may select this pixel type, but don't assume it will.
{
	return TWAIN_SetCapOneValue(ICAP_PIXELTYPE, TWTY_UINT16, (TW_UINT16)nPixType);
} // TWAIN_SetCurrentPixelType


/////////////////////////////////////////////////////////////////////////////
// ICAP_XRESOLUTION

double EZTAPI TWAIN_GetCurrentResolution(void)
{
    return TWAIN_GetCapFix32(ICAP_XRESOLUTION, 0.0);
} // TWAIN_GetCurrentResolution

double EZTAPI TWAIN_GetYResolution(void)
// Returns the current vertical resolution, in dots per *current unit*.
// In the event of failure, returns 0.0.
{
    return TWAIN_GetCapFix32(ICAP_YRESOLUTION, 0.0);
} // TWAIN_GetYResolution


int EZTAPI TWAIN_SetCurrentResolution(double dRes)
// Negotiate the current resolution for acquisition.
// Negotiation is only allowed in State 4 (TWAIN_SOURCE_OPEN)
// The source may select this resolution, but don't assume it will.
{
    return TWAIN_SetXResolution(dRes) &&
           TWAIN_SetYResolution(dRes);
} // TWAIN_SetCurrentResolution

int EZTAPI TWAIN_SetXResolution(double dRes)
{
    dxResolution = dRes;
    if (nState >= TWAIN_SOURCE_OPEN && dxResolution >= 0.0) {
        return TWAIN_SetCapFix32(ICAP_XRESOLUTION, dxResolution);
    } else {
        // wait for a better time
        return TRUE;
    }
} // TWAIN_SetXResolution

int EZTAPI TWAIN_SetYResolution(double dRes)
{
    dyResolution = dRes;
    if (nState >= TWAIN_SOURCE_OPEN && dyResolution >= 0.0) {
        return TWAIN_SetCapFix32(ICAP_YRESOLUTION, dyResolution);
    } else {
        // wait for a better time
        return TRUE;
    }
} // TWAIN_SetYResolution

/////////////////////////////////////////////////////////////////////////////
// ICAP_CONTRAST

int EZTAPI TWAIN_SetContrast(double dCon)
// (Try to) set the current contrast for acquisition.
// The TWAIN standard says that the range for this cap is -1000 ... +1000
{
	return TWAIN_SetCapFix32(ICAP_CONTRAST, dCon);
}


/////////////////////////////////////////////////////////////////////////////
// ICAP_BRIGHTNESS

int EZTAPI TWAIN_SetBrightness(double dBri)
{
	return TWAIN_SetCapFix32(ICAP_BRIGHTNESS, dBri);
}


/////////////////////////////////////////////////////////////////////////////
// ICAP_THRESHOLD

int EZTAPI TWAIN_SetThreshold(double dThresh)
{
	return TWAIN_SetCapFix32(ICAP_THRESHOLD, dThresh);
}

double EZTAPI TWAIN_GetCurrentThreshold(void)
{
    double dVal = -1;
    HCONTAINER hcon = TWAIN_GetCurrent(ICAP_THRESHOLD);
    if (hcon) {
        dVal = CONTAINER_CurrentValue(hcon);
        CONTAINER_Free(hcon);
    }
    return dVal;
} // TWAIN_GetCurrentThreshold


/////////////////////////////////////////////////////////////////////////////
// ICAP_XFERMECH

int EZTAPI TWAIN_XferMech(void)
{
	return TWAIN_GetCapUint16(ICAP_XFERMECH, XFERMECH_NATIVE);
} // TWAIN_XferMech

int EZTAPI TWAIN_SetXferMech(int mech)
{
    if (TWAIN_SetCapOneValue(ICAP_XFERMECH, TWTY_UINT16, (TW_UINT16)mech)) {
        nXferMech = mech;
        return TRUE;
    } else {
        return FALSE;
    }
} // TWAIN_SetXferMech


int EZTAPI TWAIN_SupportsFileXfer(void)
{
    HCONTAINER hcon = TWAIN_Get(ICAP_XFERMECH);
    BOOL bFile = CONTAINER_ContainsValue(hcon, XFERMECH_FILE);
    CONTAINER_Free(hcon);
    return bFile;
}

/////////////////////////////////////////////////////////////////////////////
// ICAP_SUPPORTEDSIZES
//

int EZTAPI TWAIN_SetPaperSize(int nPaper)
{
	return TWAIN_SetCapOneValue(ICAP_SUPPORTEDSIZES, TWTY_UINT16, nPaper);
}


/////////////////////////////////////////////////////////////////////////////
// Capability Negotiation: Document Feeder
//

int EZTAPI TWAIN_HasFeeder(void)
// Return TRUE(1) if it seems that the source has a document feeder.
{
    BOOL bFeeder = TWAIN_IsFeederSelected();
    BOOL bYes = bFeeder;
    if (!bFeeder) {
        if (TWAIN_SelectFeeder(TRUE)) {
            bYes = TRUE;
        }
        TWAIN_SelectFeeder(FALSE);
    }
    return bYes; 
} // TWAIN_HasFeeder


BOOL EZTAPI TWAIN_IsAutoFeedOn(void)
// Return TRUE(1) if automatic feeding is enabled, otherwise FALSE(0).
{
    return TWAIN_GetCapBool(CAP_AUTOFEED, FALSE);
} // TWAIN_IsAutoFeedOn


int EZTAPI TWAIN_SetAutoFeed(int fYes)
// (Try to) turn on/off automatic feeding thru the feeder
// This function selects the feeder if it is not already selected.
// Return TRUE(1) if successful, FALSE(0) otherwise.
{
    bAutoFeed = (fYes != 0);
    if (nState >= TWAIN_SOURCE_OPEN) {
        // Enable the feeder first - even if that fails, go ahead, in case
        // the DS is brain damaged.
        TWAIN_SelectFeeder(TRUE);
        return TWAIN_SetCapOneValue(CAP_AUTOFEED, TWTY_BOOL, (TW_BOOL)bAutoFeed);
    } else {
        return TRUE;
    }
} // TWAIN_SetAutoFeed


BOOL EZTAPI TWAIN_IsFeederSelected(void)
// Return TRUE(1) if the image source is scanning from the feeder.
{
    return TWAIN_GetCapBool(CAP_FEEDERENABLED, FALSE);
} // TWAIN_IsFeederSelected


int EZTAPI TWAIN_SelectFeeder(int bFeeder)
// (Try to) select or deselect the feeder
// Return TRUE(1) if successful, FALSE(0) otherwise.
{
    bFeeder = (bFeeder != 0);
    return TWAIN_IsFeederSelected() == bFeeder ||
           TWAIN_SetCapOneValue(CAP_FEEDERENABLED, TWTY_BOOL, (TW_BOOL)bFeeder);
} // TWAIN_SelectFeeder


BOOL EZTAPI TWAIN_IsFeederLoaded(void)
{
    return TWAIN_GetCapBool(CAP_FEEDERLOADED, FALSE);
} // TWAIN_IsFeederLoaded


/////////////////////////////////////////////////////////////////////////////
// Duplex Scanning

int EZTAPI TWAIN_GetDuplexSupport(void)
// Query the device for duplex scanning support.
//   Return values:
// 0    = no support (or error, or query not recognized)
// 1    = 1-pass duplex
// 2    = 2-pass duplex
{
    return TWAIN_GetCapUint16(CAP_DUPLEX, 0);
} // TWAIN_GetDuplexSupport


int EZTAPI TWAIN_EnableDuplex(int fYes)
// Enable (fYes=1) or disable (fYes=0) duplex scanning.
// Returns TRUE(1) if successful, FALSE(0) otherwise.
{
    bDuplex = (fYes != 0);
    if (nState >= TWAIN_SOURCE_OPEN) {
        return TWAIN_SetCapOneValue(CAP_DUPLEXENABLED, TWTY_BOOL, (TW_BOOL)bDuplex);
    } else {
        // wait for source open
        return TRUE;
    }
} // TWAIN_EnableDuplex


BOOL EZTAPI TWAIN_IsDuplexEnabled(void)
// Returns TRUE(1) if the device supports duplex scanning
// and duplex scanning is enabled.  FALSE(0) otherwise.
{
    return TWAIN_GetCapBool(CAP_DUPLEXENABLED, FALSE);
} // TWAIN_IsDuplexEnabled

/////////////////////////////////////////////////////////////////////////////
// CAP_UICONTROLLABLE

int EZTAPI TWAIN_HasControllableUI(void)
// Return 1 if source claims UI can be hidden (see SetHideUI above)
// Return 0 if source says UI *cannot* be hidden
// Return -1 if source (pre TWAIN 1.6) cannot answer the question.
{
   TW_BOOL bCUI = FALSE;
   if (TWAIN_GetCapCurrent(CAP_UICONTROLLABLE, TWTY_BOOL, &bCUI)) {
       return (int)bCUI;
   }
   return -1;
} // TWAIN_HasControllableUI

/////////////////////////////////////////////////////////////////////////////
// CAP_INDICATORS

int EZTAPI TWAIN_SetIndicators(BOOL bVisible)
// Tell the source to show (hide) progress indicators during acquisition.
{
    bIndicators = (bVisible != 0);
    if (nState >= TWAIN_SOURCE_OPEN) {
        return TWAIN_SetCapOneValue(CAP_INDICATORS, TWTY_BOOL, (TW_BOOL)bIndicators);
    } else {
        // Defer negotiation until source is open
        return TRUE;
    }
} // TWAIN_SetIndicators

/////////////////////////////////////////////////////////////////////////////
// Capability Negotiation
// ICAP_TILES

BOOL EZTAPI TWAIN_Tiled(void)
{
   return TWAIN_GetCapBool(ICAP_TILES, FALSE);
} // TWAIN_Tiled

int EZTAPI TWAIN_SetTiled(BOOL bTiled)
{
    bTiled = (bTiled != 0);
    return TWAIN_Tiled() == bTiled ||
        TWAIN_SetCapOneValue(ICAP_TILES, TWTY_BOOL, (TW_BOOL)bTiled);
} // TWAIN_SetTiled


/////////////////////////////////////////////////////////////////////////////
// ICAP_PLANARCHUNKY

int EZTAPI TWAIN_PlanarChunky(void)
{
	return TWAIN_GetCapUint16(ICAP_PLANARCHUNKY, -1);
} // TWAIN_PlanarChunky

int EZTAPI TWAIN_SetPlanarChunky(int shape)
{
   return TWAIN_SetCapOneValue(ICAP_PLANARCHUNKY, TWTY_UINT16, (TW_UINT16)shape);
} // TWAIN_SetPlanarChunky


/////////////////////////////////////////////////////////////////////////////
// ICAP_PIXELFLAVOR

int EZTAPI TWAIN_PixelFlavor(void)
{
	return TWAIN_GetCapUint16(ICAP_PIXELFLAVOR, -1);
} // TWAIN_PixelFlavor

int EZTAPI TWAIN_SetPixelFlavor(int flavor)
{
   return TWAIN_SetCapOneValue(ICAP_PIXELFLAVOR, TWTY_UINT16, (TW_UINT16)flavor);
} // TWAIN_SetPixelFlavor


/////////////////////////////////////////////////////////////////////////////
// ICAP_COMPRESSION

int EZTAPI TWAIN_Compression(void)
{
	return TWAIN_GetCapUint16(ICAP_COMPRESSION, TWCP_NONE);
} // TWAIN_Compression

int EZTAPI TWAIN_SetCompression(int compression)
{
   return TWAIN_SetCapOneValue(ICAP_COMPRESSION, TWTY_UINT16, (TW_UINT16)compression);
} // TWAIN_SetCompression


/////////////////////////////////////////////////////////////////////////////
// ICAP_AUTOBRIGHT

int EZTAPI TWAIN_SetAutoBright(BOOL bOn)
{
    bOn = (bOn != 0);
    return TWAIN_SetCapOneValue(ICAP_AUTOBRIGHT, TWTY_BOOL, (TW_BOOL)bOn);
} // TWAIN_SetAutoBright


/////////////////////////////////////////////////////////////////////////////
// ICAP_LIGHTPATH

int EZTAPI TWAIN_SetLightPath(BOOL bTransmissive)
{
    bTransmissive = (bTransmissive != 0);
    return TWAIN_SetCapOneValue(ICAP_LIGHTPATH, TWTY_UINT16, (TW_UINT16)bTransmissive);
} // TWAIN_SetLightPath


/////////////////////////////////////////////////////////////////////////////
// ICAP_GAMMA

int EZTAPI TWAIN_SetGamma(double dGamma)
{
    return TWAIN_SetCapFix32(ICAP_GAMMA, dGamma);
} // TWAIN_SetGamma

/////////////////////////////////////////////////////////////////////////////
// ICAP_SHADOW

int EZTAPI TWAIN_SetShadow(double d)
{
    return TWAIN_SetCapFix32(ICAP_SHADOW, d);
} // TWAIN_SetShadow

/////////////////////////////////////////////////////////////////////////////
// ICAP_HIGHLIGHT

int EZTAPI TWAIN_SetHighlight(double d)
{
    return TWAIN_SetCapFix32(ICAP_HIGHLIGHT, d);
} // TWAIN_SetHighlight


/////////////////////////////////////////////////////////////////////////////
// DAT_IMAGELAYOUT

int EZTAPI TWAIN_SetImageLayout(double L, double T, double R, double B)
{
    TW_IMAGELAYOUT layout;
    if (TWAIN_State() != TWAIN_SOURCE_OPEN) {
        return RecordError(EZTEC_NOT_STATE_4);
    }
    
    SetFix32(&layout.Frame.Left, L);
    SetFix32(&layout.Frame.Top, T);
    SetFix32(&layout.Frame.Right, R);
    SetFix32(&layout.Frame.Bottom, B);
    
    layout.DocumentNumber = TWON_DONTCARE32;
    layout.PageNumber = TWON_DONTCARE32;
    layout.FrameNumber = TWON_DONTCARE32;
    return TWAIN_DS(DG_IMAGE, DAT_IMAGELAYOUT, MSG_SET, &layout);
} // TWAIN_SetImageLayout


int EZTAPI TWAIN_GetImageLayout(double *L, double *T, double *R, double *B)
{
    TW_IMAGELAYOUT layout;
    if (0 == TWAIN_DS(DG_IMAGE, DAT_IMAGELAYOUT, MSG_GET, &layout)) {
        return 0;
    }
    *L = Fix32ToFloat(layout.Frame.Left);
    *T = Fix32ToFloat(layout.Frame.Top);
    *R = Fix32ToFloat(layout.Frame.Right);
    *B = Fix32ToFloat(layout.Frame.Bottom);
    return 1;    
} // TWAIN_GetImageLayout


int EZTAPI TWAIN_GetDefaultImageLayout(double *L, double *T, double *R, double *B)
{
    TW_IMAGELAYOUT layout;
    if (0 == TWAIN_DS(DG_IMAGE, DAT_IMAGELAYOUT, MSG_GETDEFAULT, &layout)) {
        return 0;
    }
    *L = Fix32ToFloat(layout.Frame.Left);
    *T = Fix32ToFloat(layout.Frame.Top);
    *R = Fix32ToFloat(layout.Frame.Right);
    *B = Fix32ToFloat(layout.Frame.Bottom);
    return 1;    
} // TWAIN_GetDefaultImageLayout


int EZTAPI TWAIN_ResetImageLayout(void)
{
    TW_IMAGELAYOUT layout;
    return TWAIN_DS(DG_IMAGE, DAT_IMAGELAYOUT, MSG_RESET, &layout);
} // TWAIN_ResetImageLayout


/////////////////////////////////////////////////////////////////////////////
// TW_FIX32 conversions

long EZTAPI TWAIN_ToFix32R(int Numerator, int Denominator)
{
    return TWAIN_ToFix32((double)Numerator / Denominator);
} // TWAIN_ToFix32R


long EZTAPI TWAIN_ToFix32(double r)
{
	TW_FIX32 fix;
	assert(sizeof (TW_FIX32) == sizeof (long));
    SetFix32(&fix, r);
	return *(long*)&fix;
} // TWAIN_ToFix32


double EZTAPI TWAIN_Fix32ToFloat(long nfix)
{
   TW_FIX32 fix;
   TW_INT32 val;
   *(long*)&fix = nfix;
	val = ((TW_INT32)fix.Whole << 16) | ((TW_UINT32)fix.Frac & 0xffff);
	return val / 65536.0;
} // TWAIN_Fix32ToFloat


/////////////////////////////////////////////////////////////////////////////
// Low-Level Capability Functions
/////////////////////////////////////////////////////////////////////////////

int EZTAPI TWAIN_SetCapFix32R(unsigned Cap, int Numerator, int Denominator)
{
    return TWAIN_SetCapFix32(Cap, (double)Numerator / Denominator);
} // TWAIN_SetCapFix32R


int EZTAPI TWAIN_SetCapFix32(unsigned Cap, double dVal)
{
	BOOL			bSuccess = FALSE;

    HCONTAINER hcon = CONTAINER_OneValue(TWTY_FIX32, dVal);
    if (!hcon) {
        TWAIN_ErrorBox(szInsuffMem);
    } else {
        bSuccess = TWAIN_Set(Cap, hcon);
        CONTAINER_Free(hcon);
    }
    return bSuccess;
} // TWAIN_SetCapFix32


int EZTAPI TWAIN_SetCapOneValue(unsigned Cap, unsigned ItemType, long ItemVal)
{
    BOOL       bSuccess = FALSE;
    HCONTAINER hcon;
    
    if (ItemType >= TWTY_FIX32) {
        TWAIN_ErrorBox(szBadType);
        return FALSE;
    }
    hcon = CONTAINER_OneValue(ItemType, ItemVal);
    if (!hcon) {
        TWAIN_ErrorBox(szInsuffMem);
        return FALSE;
    }
    bSuccess = TWAIN_Set(Cap, hcon);
    CONTAINER_Free(hcon);
    return bSuccess;
} // TWAIN_SetCapOneValue


// helper function:
int TypeMatch(unsigned nTypeA, unsigned nTypeB)
{
	// Integral types match if they are the same size.
	// All other types match only if they are equal
	return nTypeA == nTypeB ||
		   (nTypeA <= TWTY_UINT32 &&
		    nTypeB <= TWTY_UINT32 &&
		    nTypeSize[nTypeA] == nTypeSize[nTypeB]);
} // TypeMatch


static TW_UINT32 Float2Value(double dV, int twty)
{
    switch (twty) {
    case TWTY_INT8:
        return (TW_INT8)dV;
    case TWTY_INT16:
        return (TW_INT16)dV;
    case TWTY_INT32:
        return (TW_INT32)dV;
    case TWTY_UINT8:
        return (TW_UINT8)dV;
    case TWTY_UINT16:
        return (TW_UINT16)dV;
    case TWTY_UINT32:
        return (TW_UINT32)dV;
    case TWTY_BOOL:
        return (TW_BOOL)dV;
    case TWTY_FIX32:
        return TWAIN_ToFix32(dV);
    default:
        return 0;
    } // switch
} // Float2Value

static int StoreItem(void* pv, int twty, double dV)
// Store the value dV into the item at pv which has type twty
{
    switch (twty) {
    case TWTY_INT8:
        *(TW_INT8*)pv = (TW_INT8)dV;
        break;
    case TWTY_INT16:
        *(TW_INT16*)pv = (TW_INT16)dV;
        break;
    case TWTY_INT32:
        *(TW_INT32*)pv = (TW_INT32)dV;
        break;
    case TWTY_UINT8:
        *(TW_UINT8*)pv = (TW_UINT8)dV;
        break;
    case TWTY_UINT16:
        *(TW_UINT16*)pv = (TW_UINT16)dV;
        break;
    case TWTY_UINT32:
        *(TW_UINT32*)pv = (TW_UINT32)dV;
        break;
    case TWTY_BOOL:
        *(TW_BOOL*)pv = (TW_BOOL)dV;
        break;
    case TWTY_FIX32:
        SetFix32(pv, dV);
        break;
    default:
        return FALSE;
    } // switch
    return TRUE;
} // StoreItem


static double Value2Float(TW_UINT32 u32, TW_UINT16 twty)
{
    switch (twty) {
    case TWTY_INT8:
        return (TW_INT8)u32;
    case TWTY_INT16:
        return (TW_INT16)u32;
    case TWTY_INT32:
        return (TW_INT32)u32;
    case TWTY_UINT8:
        return (TW_UINT8)u32;
    case TWTY_UINT16:
        return (TW_UINT16)u32;
    case TWTY_UINT32:
        return (TW_UINT32)u32;
    case TWTY_BOOL:
        return (TW_BOOL)u32;
    case TWTY_FIX32:
        return TWAIN_Fix32ToFloat(u32);
    default:
        return 0.0;
    } // switch
} // Value2Float


static double Item2Float(void* pv, TW_UINT16 twty)
{
    switch (twty) {
    case TWTY_INT8:
        return *(TW_INT8*)pv;
    case TWTY_INT16:
        return *(TW_INT16*)pv;
    case TWTY_INT32:
        return *(TW_INT32*)pv;
    case TWTY_UINT8:
        return *(TW_UINT8*)pv;
    case TWTY_UINT16:
        return *(TW_UINT16*)pv;
    case TWTY_UINT32:
        return *(TW_UINT32*)pv;
    case TWTY_BOOL:
        return *(TW_BOOL*)pv;
    case TWTY_FIX32:
        return Fix32ToFloat(*(TW_FIX32*)pv);
    default:
        return 0.0;
    } // switch
} // Item2Float


int EZTAPI TWAIN_Reset(unsigned uCap)
{
    TW_CAPABILITY 	cap;
	if (nState < TWAIN_SOURCE_OPEN) {
		TWAIN_ErrorBox("TWAIN_Reset below State 4.");
		return 0;
	}

	// Fill in a capability structure
	cap.Cap = (TW_UINT16)uCap;					// capability id
    return TWAIN_DS(DG_CONTROL, DAT_CAPABILITY, MSG_RESET, (TW_MEMREF)&cap);
} // TWAIN_Reset


BOOL EZTAPI TWAIN_GetCapBool(unsigned cap, BOOL bValue)
{
    HCONTAINER hcon =  TWAIN_Get(cap);
    if (hcon) {
        bValue = CONTAINER_CurrentValue(hcon) != 0;
        CONTAINER_Free(hcon);
    }
    return bValue;
} // TWAIN_GetCapBool


double EZTAPI TWAIN_GetCapFix32(unsigned uCap, double dDefault)
{
	TW_FIX32 fix;
    if (!TWAIN_GetCapCurrent(uCap, TWTY_FIX32, &fix)) {
        return dDefault;
    }
	return Fix32ToFloat(fix);
} // TWAIN_GetCapFix32


int EZTAPI TWAIN_GetCapUint16(unsigned cap, int nDefault)
{
    TW_UINT16 nValue = (TW_UINT16)nDefault;
    if (!TWAIN_GetCapCurrent(cap, TWTY_UINT16, &nValue)) {
        nValue = (TW_UINT16)nDefault;
    }
    return nValue;
} // TWAIN_GetCapUint16


int EZTAPI TWAIN_GetCapCurrent(unsigned Cap, unsigned ItemType, void FAR *pVal)
{
	TW_CAPABILITY 	cap;
	void far *		pv = NULL;
	BOOL			bSuccess = FALSE;
    int             nItemSize = nTypeSize[ItemType];

	assert(pVal != NULL);

	if (nState < TWAIN_SOURCE_OPEN) {
		TWAIN_ErrorBox("Attempt to get capability value below State 4.");
		return FALSE;
	}

	// Fill in capability structure
	cap.Cap = (TW_UINT16)Cap;					// capability id
	cap.ConType = TWON_ONEVALUE;	// favorite type of container (should be ignored...)
	cap.hContainer = NULL;

	if (TWAIN_DS(DG_CONTROL, DAT_CAPABILITY, MSG_GETCURRENT, (TW_MEMREF)&cap) &&
	    cap.hContainer &&
	    (pv = GlobalLock(cap.hContainer)) != NULL) {

		if (cap.ConType == TWON_ENUMERATION) {
			TW_ENUMERATION far *pcon = (TW_ENUMERATION far *)pv;
			TW_UINT32 index = pcon->CurrentIndex;
			if (index < pcon->NumItems && TypeMatch(pcon->ItemType, ItemType)) {
				LPSTR pitem = (LPSTR)pcon->ItemList + index*nItemSize;
				FMEMCPY(pVal, pitem, nItemSize);
				bSuccess = TRUE;
			}
		} else if (cap.ConType == TWON_ONEVALUE) {
			TW_ONEVALUE far *pcon = (TW_ONEVALUE far *)pv;
			if (TypeMatch(pcon->ItemType, ItemType)) {
				FMEMCPY(pVal, &pcon->Item, nItemSize);
				bSuccess = TRUE;
			}
        } else if (cap.ConType == TWON_RANGE) {
            TW_RANGE far *pcon = (TW_RANGE far *)pv;
            if (TypeMatch(pcon->ItemType, ItemType)) {
                FMEMCPY(pVal, &pcon->CurrentValue, nItemSize);
                bSuccess = TRUE;
            }
        }
        
    }
	
	if (pv) GlobalUnlock(cap.hContainer);
	if (cap.hContainer) GlobalFree(cap.hContainer);

	return bSuccess;
} 

//-------------------------- The primitive functions


int EZTAPI TWAIN_DS(unsigned long dg, unsigned dat, unsigned msg, void FAR *pd)
// Call the current source with a triplet
{
   int bOk = FALSE;
   assert(nState >= TWAIN_SOURCE_OPEN);
   rc = TWRC_FAILURE;
   if (dg == DG_IMAGE) {
      if (dat == DAT_IMAGEMEMXFER) {
         if (msg == MSG_GET && pd != NULL) {
            pTW_IMAGEMEMXFER pmxb = (pTW_IMAGEMEMXFER)pd;
            pmxb->Compression   = TWON_DONTCARE16;
            pmxb->BytesPerRow   = TWON_DONTCARE32;
            pmxb->Columns       = TWON_DONTCARE32;
            pmxb->Rows          = TWON_DONTCARE32;
            pmxb->XOffset       = TWON_DONTCARE32;
            pmxb->YOffset       = TWON_DONTCARE32;
            pmxb->BytesWritten  = TWON_DONTCARE32;
         }
      }
   }
   if (pDSM_Entry) {
       static nNested;

#ifdef TRACE
       char szLog[256];
       sprintf(szLog, "TWAIN_DS(%u, %u, %u)", dg, dat, msg);
       OutputDebugString(szLog);
#endif
       if (pfsLog) {
          if (dg != DG_CONTROL || dat != DAT_EVENT || msg != MSG_PROCESSEVENT) {
             fprintf(pfsLog, "TWAIN_DS(%u, %u, %u)\n", dg, dat, msg);
             fflush(pfsLog);
          }
       }

       //assert(nNested == 0);
       if (nNested != 0) {
           // nested call, refuse it
           OutputDebugString(" - blocked\n");
           //return FALSE;
       }
       nNested++;
      rc = (*pDSM_Entry)(&AppId, &SourceId,
						   (TW_UINT32)dg,
                     (TW_UINT16)dat,
                     (TW_UINT16)msg,
                     (TW_MEMREF)pd);
      nNested--;
#ifdef TRACE
       OutputDebugString("\n");
#endif
      bOk = (rc == TWRC_SUCCESS);

      if (dg == DG_CONTROL) {
          switch (dat) {
          case DAT_EVENT:
              if (msg == MSG_PROCESSEVENT) {
                  if (((pTW_EVENT)pd)->TWMessage == MSG_XFERREADY) {
                      SetState(TWAIN_TRANSFER_READY);
                  }
                  bOk = (rc == TWRC_DSEVENT);
              }
              break;
              
          case DAT_CAPABILITY:
              break;

          case DAT_PENDINGXFERS:
              if (msg == MSG_ENDXFER && bOk) {
                  SetState(((pTW_PENDINGXFERS)pd)->Count ? TWAIN_TRANSFER_READY
                      : TWAIN_SOURCE_ENABLED);
              }
              if (msg == MSG_RESET && bOk) {
                  SetState(TWAIN_SOURCE_ENABLED);
              }
              break;
          case DAT_USERINTERFACE:
              if (msg == MSG_ENABLEDS) {
                  if (rc == TWRC_FAILURE) {
                      RecordError(EZTEC_DS_FAILURE);
                  } else {
                      // rc could be either SUCCESS or CHECKSTATUS
                      SetState(TWAIN_SOURCE_ENABLED);
                      bOk = TRUE;
                  }
              }
              if (msg == MSG_DISABLEDS && bOk) {
                  SetState(TWAIN_SOURCE_OPEN);
              }
              break;

          case DAT_SETUPMEMXFER:
              if (msg == MSG_GET && bOk) {
                  nMemBuffer = 0;
              }
              break;
          } // switch
          
      } else if (dg == DG_IMAGE) {
         if (dat == DAT_IMAGENATIVEXFER || dat == DAT_IMAGEFILEXFER) {
            // Native and File transfers work much the same way.
            if (msg == MSG_GET) {
               bOk = (rc == TWRC_XFERDONE);
               switch (rc) {
               case TWRC_XFERDONE:
               case TWRC_CANCEL:
                  SetState(TWAIN_TRANSFERRING);
                  // Need to acknowledge end of transfer with
                  // DG_CONTROL / DAT_PENDINGXFERS / MSG_ENDXFER
                  break;
                  
               case TWRC_FAILURE:
               default:
                  // Transfer failed (e.g. insufficient memory, write-locked file)
                  // check condition code for more info
                  SetState(TWAIN_TRANSFER_READY);
                  // The image is still pending
                  break;
               } // switch
            }
         } else if (dat == DAT_IMAGEMEMXFER) {
            if (msg == MSG_GET) {
               bOk = FALSE;
               switch (rc) {
               case TWRC_SUCCESS:
               case TWRC_XFERDONE:
                  bOk = TRUE;
                  nMemBuffer++;
                  SetState(TWAIN_TRANSFERRING);
                  break;
               case TWRC_FAILURE:
                  // "If the failure occurred during the transfer of the first
                  // buffer, the session is in State 6. If the failure occurred
                  // on a subsequent buffer, the session is in State 7."
                  SetState(nMemBuffer == 0 ? TWAIN_TRANSFER_READY : TWAIN_TRANSFERRING);
                  break;
               case TWRC_CANCEL:
                  // Transfer cancelled, no state change.
                  TWAIN_BreakModalLoop();
                  break;
               } // switch
            }
         }
      }
   }
   return bOk;
} // TWAIN_DS



int EZTAPI TWAIN_Mgr(unsigned long dg, unsigned dat, unsigned msg, void FAR *pd)
// Call the Source Manager with a triplet
{
   int bOk = FALSE;
	rc = TWRC_FAILURE;
	if (pDSM_Entry) {
		rc = (*pDSM_Entry)(&AppId, NULL,
						   (TW_UINT32)dg,
						   (TW_UINT16)dat,
						   (TW_UINT16)msg,
						   (TW_MEMREF)pd);
   	bOk = (rc == TWRC_SUCCESS);
      if (dg == DG_CONTROL) {
         if (dat == DAT_IDENTITY) {
            if (msg == MSG_OPENDS) {
               if (bOk) {
                  // Source opened - record identity for future triplets
                  FMEMCPY(&SourceId, pd, sizeof (TW_IDENTITY));
                  SetState(TWAIN_SOURCE_OPEN);
                  if (pfsLog) {
                     fprintf(pfsLog,
                        "Product \"%s\"\n"
                        "Version %u-%u  Vendor: \"%s\"\n",
                        (char*)SourceId.ProductName,
                        (unsigned)SourceId.Version.MajorNum,
                        (unsigned)SourceId.Version.MinorNum,
                        (char*)SourceId.Manufacturer);
                  }
               } else {
            		RecordError(EZTEC_DSM_FAILURE);
               }
            }
            if (msg == MSG_CLOSEDS && bOk) {
               SetState(TWAIN_SM_OPEN);
            }
         }
         if (dat == DAT_PARENT) {
            if (msg == MSG_OPENDSM && bOk) {
               SetState(TWAIN_SM_OPEN);
            }
            if (msg == MSG_CLOSEDSM && bOk) {
               SetState(TWAIN_SM_LOADED);
            }
         }
      }
	}
	return bOk;
} // TWAIN_Mgr


// Error Diagnosis & Reporting

unsigned EZTAPI TWAIN_GetResultCode(void)
{
	return rc;
} // TWAIN_GetResultCode



unsigned EZTAPI TWAIN_GetConditionCode(void)
{
	TW_STATUS	twStatus;
   TW_INT16    rcLast = rc;

	twStatus.ConditionCode = TWCC_BUMMER;			// what can I say. 

	if (nState >= 4) {
		// get source status if open
		TWAIN_DS(DG_CONTROL, DAT_STATUS, MSG_GET, (TW_MEMREF)&twStatus);
	} else if (nState == 3) {
		// otherwise get source manager status
		TWAIN_Mgr(DG_CONTROL, DAT_STATUS, MSG_GET, (TW_MEMREF)&twStatus);
	}
   rc = rcLast;
   return twStatus.ConditionCode;
} // TWAIN_GetConditionCode


int EZTAPI TWAIN_UserClosedSource(void)
{
    return bCloseDsRequested;
} // TWAIN_UserClosedSource


void EZTAPI TWAIN_ReportLastError(LPCSTR pzMsg)
{
    char szMsg[256] = "";			// scratch buffer for messages
    if (pzMsg) {
        lstrcpy(szMsg, pzMsg);
    }
    if (szMsg[0]) {
        lstrcat(szMsg, "\n");
    }
    lstrcat(szMsg, pszErrDescrip[nErrDetail]);
    if (nErrDetail > EZTEC_START_TRIPLET_ERRS && nErrDetail < EZTEC_END_TRIPLET_ERRS) {
        wsprintf(szMsg+strlen(szMsg), "\nRC:   %s\nCC:   %s",
            (LPSTR)pszRC[nErrRC], (LPSTR)pszCC[nErrCC]);
    } else {
    }
    if (nErrRC == TWRC_FAILURE && nErrCC == TWCC_OPERATIONERROR) {
        // don't report this to user - DS has already reported it.
#ifdef _DEBUG
        OutputDebugString("EZTW: TWRC_FAILURE, TWCC_OPERATIONERROR\n");
#endif
    } else {
        TWAIN_ErrorBox(szMsg);
    }
} // ReportLastError


int EZTAPI TWAIN_SuppressErrorMessages(int nSuppress)
{
    int nPrev = bNoMessages;
    bNoMessages = (nSuppress != 0);
    return nPrev;
} // TWAIN_SuppressErrorMessages


void EZTAPI TWAIN_ErrorBox(LPCSTR pzMsg)
{
   if (bNoMessages) {
#ifdef _DEBUG
      OutputDebugString("** EZTWAIN Error (Suppressed!) **\n** ");
      OutputDebugString(pzMsg);
      OutputDebugString("\n");
#endif
      if (pfsLog) {
         fprintf(pfsLog, "TWAIN_ErrorBox ** SUPPRESSED **\nMsg: %s\n", pzMsg);
         fflush(pfsLog);
      }
   } else {
#ifdef _DEBUG
      OutputDebugString("** EZTWAIN Error **\n** ");
      OutputDebugString(pzMsg);
      OutputDebugString("\n");
#endif
      if (pfsLog) {
         fprintf(pfsLog, "TWAIN_ErrorBox\nMsg: %s\n", pzMsg);
         fflush(pfsLog);
      }
      MessageBox(NULL, pzMsg, "TWAIN Error", MB_ICONINFORMATION | MB_OK);
   }
} // TWAIN_ErrorBox


//------------ Private functions

void SetState(int n)
{
#ifdef _DEBUG
   char szMsg[256];
   wsprintf(szMsg, "EZTW:State %d\n", n);
   OutputDebugString(szMsg);
#endif
   if (pfsLog && n != nState) {
      fprintf(pfsLog, "STATE %d -> %d\n", nState, n);
      fflush(pfsLog);
   }
   nState = n;
} // SetState


HWND GetValidHwnd(HWND hwnd)
// Returns a valid window handle as follows:
// If hwnd is a valid window handle, hwnd is returned.
// Otherwise a proxy window handle is created and returned.
// Once created, a proxy window handle is destroyed when
// the source manager is unloaded.
// If hwnd is an invalid window handle (other than NULL)
// an error box is displayed.
{
    if (!IsWindow(hwnd)) {
        if (hwnd != NULL) {
            TWAIN_ErrorBox("EZTWAIN: window handle is invalid");
            hwnd = NULL;
        }
        if (!hwndProxy) {
            hwndProxy = CreateProxyWindow();
            if (!IsWindow(hwndProxy)) {
                TWAIN_ErrorBox("EZTWAIN: Unable to create proxy window");
                hwndProxy = NULL;
            }
        }
        hwnd = hwndProxy;
    }
    return hwnd;
} // GetValidHwnd


HWND CreateProxyWindow(void)
{
	HWND hwnd;
	hwnd = CreateWindow("STATIC",                // class
						"Acquire Proxy",              // title
						WS_POPUPWINDOW,               // style
						CW_USEDEFAULT, CW_USEDEFAULT, // x, y
						CW_USEDEFAULT, CW_USEDEFAULT, // width, height
						HWND_DESKTOP,                 // parent window
						NULL,                         // hmenu
						hinstLib,                     // hinst
						NULL);                        // lpvparam
	return hwnd;
} // CreateProxyWindow


int RecordError(EZTWAIN_ERROR_CODE ed)
{
	if (nErrDetail == EZTEC_NONE) {
		nErrDetail = ed;
		if (ed > EZTEC_START_TRIPLET_ERRS && ed < EZTEC_END_TRIPLET_ERRS) {
			nErrRC = TWAIN_GetResultCode();
			nErrCC = TWAIN_GetConditionCode();
		} else {
			nErrRC = 0;
			nErrCC = 0;
		}
	}
	return FALSE;
} // RecordError


int EZTAPI TWAIN_LastErrorCode(void)
// Return the last internal EZTWAIN error code.
{
    return (int)nErrDetail;
} // TWAIN_LastErrorCode


void EZTAPI TWAIN_ClearError(void)
// Set the EZTWAIN internal error code to EZTEC_NONE
{
	nErrDetail = EZTEC_NONE;
} // TWAIN_ClearError


void EZTAPI TWAIN_LogFile(int fLog)
{
   if (fLog != 0) {
      if (!pfsLog) {
         // create an empty log file in text mode
         pfsLog = fopen( "c:\\eztwain.log", "wt" );
         if (!pfsLog) {
            return;
         }
         setvbuf(pfsLog, NULL, _IONBF, 0);
         fprintf(pfsLog, "-- Start of EZTWAIN log\n");
         fprintf(pfsLog, "EZTWAIN version %d.%02d\n", VERSION / 100, VERSION % 100);
         
         {
            time_t ltime;
            
            _tzset();
            time(&ltime);
            fprintf(pfsLog, "Universal time: %s", asctime(gmtime(&ltime)));
            fprintf(pfsLog, "Local time:     %s\n", asctime(localtime(&ltime)));
         }
         fflush(pfsLog);
      }
   } else {
      // Turn off logging
      if (pfsLog) {
         fprintf(pfsLog, "-- End of EZTWAIN log\n");
         fclose(pfsLog);
         pfsLog = NULL;
      }
   }
} // TWAIN_Logging

TW_UINT32 Intersect(TW_UINT32 wMask, TW_UINT32 nItems, TW_UINT16 far *pItem)
{
	TW_UINT32	wSet = 0;
	TW_UINT16	i;

	// In wSet, construct set of available items.
	// Note that items that cannot be represented in wMask are also
	// unrepresentable in wSet so are implicitly discarded
	for (i = 0 ; i < nItems; i++) {
		wSet |= 1 << pItem[i];
	} // for

	// Discard anything in wMask that isn't in wSet
	wMask &= wSet;

	// Re-fill the item table with intersection set
	for (nItems = i = 0; wMask ; wMask>>=1,i++) {
		if (wMask & 1) {
			pItem[nItems++] = i;
		}
	} // for

	return nItems;
} // Intersect


unsigned FindIndex16(TW_UINT32 nItems, TW_UINT16 far *plist, TW_UINT16 uVal, unsigned nDefault)
// Find the index of uVal in the list plist which contains nItems entries.
// If not found, return nDefault.
{
   unsigned i;
   for (i = 0; i < nItems; i++) {
      if (plist[i] == uVal) {
         return i;
      }
   }
   return nDefault;
} // FindIndex16


double Fix32ToFloat(TW_FIX32 fix)
{
	TW_INT32 val = ((TW_INT32)fix.Whole << 16) | ((TW_UINT32)fix.Frac & 0xffff);
	return val / 65536.0;
} // Fix32ToFloat

void SetFix32(TW_FIX32* pfix, double r)
{
   // Note: This round-to-nearest is new in TWAIN 1.7
	TW_INT32 val = EZround(r * 65536.0);
	pfix->Whole = (TW_INT16)(val >> 16);		// most significant 16 bits
	pfix->Frac = (TW_UINT16)(val & 0xffff);		// least
} // SetFix32


long EZround(double d)
// Round to nearest integral value
{
   // Note: ANSI C converts float to int by truncating toward 0.
    return (long)((d >= 0.0) ? d + 0.5 : d - 0.5);
} // EZround


/*///////////////////////////////////////////////////////////////////////
// History

 2.70 2001-03-05 spike
      Fix:   PreAcquire was doing EnableDuplex all the time - crashed the
             Canon PowerShot DS.  Changed logic to only negotiate if client
             app has explicitly requested.  Ditto for AutoFeed.

 2.66 00.10.24 spike
      Fix:   Changed TWAIN_Acquire to use native transfer by default.

 2.65 00.10.06 spike
      Fix:   So many problems with memory transfer, I'm about to give up
             and go back to using native as the primary transfer mode!
      Fix:   GetCurrentThreshold is more forgiving, should accept almost
             any plausible returned container. (per req. David Offen)

 2.64 00.09.19 spike
      Delta: Moved container stuff into contain.c, DIB stuff into ezdib.c
      New:   DIB_SwapRedBlue, DIB_SetColorCount, DIB_PutOnClipboard
      New:   DIB_WriteToBmpFile, DIB_WriteRowChannel, DIB_GetFromClipboard
      New:   DIB_Copy, DIB_ColorTableR, DIB_ColorTableG, DIB_ColorTableB
      Fix:   (I hope) round resolution values (in DPI) to the nearest integer,
             if they are "really close" already e.g. 99.998 => 100.0
      Fix:   special hack in memory transfer code to handle HP PictureScan 3.0.

 2.63 00.09.03 spike
      New:   TWAIN_SetIndicators, TWAIN_EnableDuplex
             will accept settings with no source open, and try to negotiate
             the requested settings at the appropriate later time.

 2.62 00.08.30 spike
      New:   TWAIN_SetAppTitle (per request of Dave Offen)

 2.61 00.08.26 spike
      New:   TWAIN_LogFile to turn log file on and off at runtime.
      New:   EZTWAIN_Attach, EZTWAIN_Detach, for directly-linked apps

 2.60 00.07.17 spike
      Fix:   Bug in memory-transfer code. (Thanks Francisco)

 2.59 00.07.10 spike
      New:   Handle tiled memory transfers.

 2.58 00.06.13 spike
      Fix:   Will use Intel's IJL11.DLL if present.
      New:   GetImageLayout, GetDefaultImageLayout, ResetImageLayout
      New:   SuppressErrorMessages, ClearError, LastErrorCode
      New:   When a transfer is cancelled by the user, LastErrorCode
             will return EZTEC_USER_CANCEL, in all transfer modes.
      New:   UserClosedSource

 2.57 00.05.16 spike
      Fix:   Memory transfer would complain if Source tried to transfer
             1 more row than was in image - which is not an uncommon source bug.

 2.56 00.04.24 spike
      Fix:   AcquireFile would not accept TIFF as a format. [thnx2 Peter Rettig]

 2.55 00.04.11 spike
      Fix:   Improved code in GetCapBool [thanx2 Paul Roder]
      Fix:   GetCapCurrent handles TW_RANGE container [thanx2 Dave Offen]
      New:   Much better reporting of unexpected conditions during
             memory transfer. [inspired by Kjell Petersen]

 2.53 00.03.07 spike
      Fix:   In TWAIN_UnloadSourceManager, destroy the
             proxy window, if there is one. [per Mark Kellam]
      Fix:   Handle odd case where WM_QUIT was seen (but ignored)
             inside TWAIN_EmptyMessageQueue. [Logitech QC]
      New:   TWAIN_AcquireMemoryCallback and TWAIN_DisableParent,
             based on code contributed by David Offen of Arkenstone.
             TWAIN_GetDisableParent (to go with ..DisableParent)
      New:   TWAIN_WaitForNativeXfer will now return a pending
             DIB if there is one.
             TWAIN_IsTransferReady - tests to see if source has
             an image ready to transfer OR if there is a DIB
             waiting from an earlier transfer.
             These two changes make it fairly easy to do asynch.
             acquires - while the application stays active - as
             long as you can hook TWAIN_MessageHook into your
             main message pump - see the latest Twerp.

 2.52 00.02.09 spike
                   Fix:   TWAIN_LoadNativeFromFile had an '=' where it
                          needed a '==' - should not have caused a problem
                          for anybody.  Thanks Hayden!
                   Fix:   Several recent functions were declared with BOOL params:
                          The Visual Basic value 'True' = -1!  So these functions
                          now have code to convert any non-zero 'BOOL' argument to 1.

 2.51 99.12.01 spike
                   Added: TWAIN_GetSourceList, TWAIN_GetNextSourceName
                          TWAIN_SourceName, TWAIN_GetSourceName
                          TWAIN_GetDefaultSourceName
                          TWAIN_OpenSource
                          TWAIN_SupportsFileXfer, TWAIN_AcquireFile

 2.50 99.11.28 spike
                   Added: TWAIN_Acquire - uses memory transfer
                          changed AcquireToClipboard, AcquireToFilename
                          to use the new TWAIN_Acquire.

 2.10 99.10.10 spike
                   Added: Support for JPEG file read & write
                          TWAIN_IsJpegAvailable
                          TWAIN_SetSaveFormat/TWAIN_GetSaveFormat
                          TWAIN_SetJpegQuality/TWAIN_GetJpegQuality
                          DIB_WriteToJpeg

 2.09 99.09.30 spike
                   Fix:   several functions were not exported by DLL
                          AcquireToFilename - return code wrong
                            in case of user cancelled save dialog.
 2.08 99.09.01 spike
                   Added:
                          TWAIN_SetCapFix32R
                          TWAIN_ToFix32R
 2.07 99.08.28 spike
                   Added: DIB_Free
                          TWAIN_GetDuplexSupport
                          TWAIN_EnableDuplex
                          TWAIN_IsDuplexEnabled
 2.06 99.06.10 spike
                   Added: DIB_ReadRowRGB
                          TWAIN_SetXResolution, TWAIN_SetYResolution
                          TWAIN_SetCapFix32
                   Fix:   TWAIN_SetCapOneValue handled FIX32 wrong
                          setting FIX32 caps to non-integral values was busted.

 2.05     06/09/99 Fix:   Some wrong return codes in CONTAINER_ functions
                          CONTAINER_SetItem didn't work on a OneValue.
                          CONTAINER_OneValue wrong for frames & strings
                   Added: CONTAINER_SelectDefaultValue
                          CONTAINER_SelectCurrentValue
                          CONTAINER_SetItemFrame
                          CONTAINER_SetItemString
 2.00 b2  01/29/99 Fix:   Return codes wrong in AcquireToFilename.
                   Added: GetCap, SetCap
                          CONTAINER_* functions
 2.00 b1  01/25/99 Requires a license for commercial use.
                   Fix:   don't rely on DS for xfermech.
                   Added: (TWAIN_ prefix omitted for brevity)
                          HasControllableUI, BuildName;
                          SetMultiTransfer, GetMultiTransfer;
                          HasFeeder, IsFeederSelected, SelectFeeder;
                          IsAutoFeedOn, SetAutoFeed;
                          SetImageLayout;
                          Compression, SetCompression;
                          Tiled, SetTiled;
                          PlanarChunky, SetPlanarChunky;
                          PixelFlavor, SetPixelFlavor;
                          SetThreshold, GetCurrentThreshold;
                          GetCapUint16, GetCapFix32;
 1.20     09/15/98 Added: TWAIN_Testing123.
 1.12     09/14/98 Added: Fix32ToFloat, allow MSG_OPENDS triplet.
                   Added: SetXferMech, XferMech.
 1.11     08/17/98 Added: ToFix32, SetContrast, SetBrightness.
                   Fix:   TWAIN_ToFix32 to round away-from-zero.
 1.09b    07/27/98 Reverted from 1.08 to 1.06 and worked forward again.
 1.06     08/21/97 correction to message hook, fixed 32-bit exports
 1.05     11/06/96 32-bit conversion
 1.04     05/03/95 Added: WriteNativeToFile, WriteNativeToFilename,
					      FreeNative, SetHideUI, GetHideUI, SetCurrentUnits,
					      GetCurrentUnits, SetCurrentResolution, SetBitDepth,
					      SetCurrentPixelType, SetCapOneValue.
 1.0a	   06/23/94 first alpha version
 0.0       05/11/94 created

// Contact Information

 Download from:    www.dosadi.com

 Support contact:  eztwain@dosadi.com

///////////////////////////////////////////////////////////////////////*/
